package controller;

import general.Gui;
import general.Phenotype;
import general.Simulation;

import java.awt.BasicStroke;
import java.awt.Color;
import java.awt.Font;
import java.awt.Graphics2D;
import java.util.ArrayList;

import options.SimulatorOptions;

import managers.ActionHandler;
import robot.LightSensor;
import robot.Robot;
import robot.SensorHandler;
import util.Util;
import actions.ActuatorAction;
import actions.BackwardAction;
import actions.ForwardAction;
import actions.ForwardLeftAction;
import actions.ForwardRightAction;
import actions.GripAction;
import actions.ReleaseGripAction;
import actions.TurnHardLeftAction;
import actions.TurnHardRightAction;
import behaviors.AssembleBehavior;
import behaviors.AssembleRecoveryBehavior;
import behaviors.Behavior;
import behaviors.CollisionAvoidanceBehavior;
import behaviors.FollowLeaderBehavior;
import behaviors.GoalConvergingBehavior;
import behaviors.HoleAvoidanceBehavior;
import behaviors.MaintainTurn;
import behaviors.StagnationRecoveryBehavior;
import behaviors.WanderBehavior;

public class ArbitrationController extends Controller {
	public final float distanceThresholdClose;
	public final float distanceThresholdFar;
	public final float lightThresholdClose;
	public final float lightThresholdFar;

	private ArrayList<Behavior> behaviors;
	private ArrayList<ActuatorAction> actions;
	private ActionHandler actionHandler;

	private Phenotype<Float> phenotype;
	private Float[] phenotypeValues;

	private float connectionTimeLeft = 0;

	public ArbitrationController(Phenotype<Float> phenotype) {
		this.phenotype = phenotype;
		if (phenotype == null) {
			phenotypeValues = new Float[] { 0.5f, 0.25f, 0.25f, 1f, 0.5f, 0.5f, 0.5f, 0.5f, 0.5f, 0.1f, 0.3f, 0.005f, 0.0002f };
		} else {
			phenotypeValues = phenotype.getPhenotypeArray();
		}
		if (SimulatorOptions.forceSelfAssembling)
			phenotypeValues[ActionHandler.ASSEMBLE] = 1f;

		distanceThresholdClose = phenotypeValues[9];
		distanceThresholdFar = phenotypeValues[10];
		lightThresholdClose = phenotypeValues[11];
		lightThresholdFar = phenotypeValues[12];

		createActions();

		behaviors = new ArrayList<Behavior>();

		behaviors.add(new GoalConvergingBehavior(this));
		behaviors.add(new WanderBehavior());
		behaviors.add(new MaintainTurn(this));
		behaviors.add(new AssembleBehavior(this));
		behaviors.add(new FollowLeaderBehavior(this));
		behaviors.add(new CollisionAvoidanceBehavior(this));
		behaviors.add(new HoleAvoidanceBehavior(this));
		behaviors.add(new StagnationRecoveryBehavior(this));
		behaviors.add(new AssembleRecoveryBehavior(this));

		actionHandler = new ActionHandler(behaviors, actions, phenotypeValues);
	}

	public void run(float elapsedTime) {
		updateLedColorIfGoalIsInSight();

		if (connectionTimeLeft > 0 && robot.isConnected()) {
			connectionTimeLeft -= Simulation.timeStep;
			robot.setWheelForces(0, 0);
			robot.getSensorHandler().getAccelerometer().resetTimeSinceMoved();
			return;
		}
		doVotedAction(elapsedTime);
	}

	private void doVotedAction(float elapsedTime) {
		actionHandler.updateBestAction(elapsedTime);
		ActuatorAction action = actionHandler.getBestAction();

		if (action != null) {
			if (action instanceof GripAction) {
				connectionTimeLeft = SimulatorOptions.connectionTime;
			}
			action.performAction(robot);
		}
	}

	private void createActions() {
		float mainSpeed = 1.0f;
		actions = new ArrayList<ActuatorAction>();
		actions.add(new TurnHardLeftAction(-mainSpeed, mainSpeed));
		actions.add(new ForwardLeftAction((float) (mainSpeed * 0.5), mainSpeed));
		actions.add(new ForwardAction(mainSpeed, mainSpeed));
		actions.add(new ForwardRightAction(mainSpeed, (float) (mainSpeed * 0.5)));
		actions.add(new TurnHardRightAction(mainSpeed, -mainSpeed));
		actions.add(new BackwardAction(-mainSpeed, -mainSpeed));
		actions.add(new GripAction());
		actions.add(new ReleaseGripAction(-mainSpeed, -mainSpeed));
	}

	public float getLightThresholdFar() {
		return lightThresholdFar;
	}

	public float getDistanceThresholdFar() {
		return distanceThresholdFar;
	}

	private void updateLedColorIfGoalIsInSight() {
		for (LightSensor l : robot.getSensorHandler().getLightSensors()) {
			boolean redColor = l.getClosestColorToSensorColor() == Color.RED;
			boolean strongEnoughLight = l.getLightValue() >= 0.8;
			if (redColor && strongEnoughLight) {
				robot.getBackLight().setColor(Color.RED);
				// robot.getFrontLight().setColor(Color.RED);
				return;
			}
		}
		if (robot.getBackLight().getColor() != Color.GREEN) {
			robot.getBackLight().setColor(Color.GREEN);
		}
		// if (robot.getFrontLight().getColor() != Color.BLUE) {
		// robot.getFrontLight().setColor(Color.BLUE);
		// }
	}

	@Override
	public void draw(Graphics2D g) {
		g.setColor(Color.BLACK);
		int fontSize = 13;
		int stringHeight = fontSize + 3;
		int boxW = 130;
		int boxH = (int) (stringHeight * 1.5f);
		int spaceH = 15;
		int headerSpace = stringHeight * 3;
		int decimalPlaces = 2;
		int[] colsStartX = new int[] { 0, 80, 150, 280, 420, 560 };
		g.setFont(new Font("Arial", Font.BOLD, fontSize));
		// Headings
		g.drawString("Situation", colsStartX[0], stringHeight);
		g.drawString("Weight", colsStartX[0], 2 * stringHeight);
		g.drawString("Behavior", colsStartX[1], stringHeight);
		g.drawString("Weight", colsStartX[1], 2 * stringHeight);
		g.drawString("Behaviors", colsStartX[2], stringHeight);
		g.drawString("Votes", colsStartX[3], stringHeight);
		g.drawString("Actions", colsStartX[4], stringHeight);
		g.drawString("Result", colsStartX[5], stringHeight);

		g.setFont(new Font("Arial", Font.PLAIN, fontSize));
		g.translate(0, headerSpace);

		// Votes
		int BestIndex = actionHandler.getActions().indexOf(actionHandler.getBestAction());
		g.setStroke(Util.thinLine);

		float[] behaviorWeights = actionHandler.getBehaviorWeights();
		float maxWeight = 0;
		for (int i = 0; i < behaviorWeights.length; i++) {
			if (behaviorWeights[i] > maxWeight)
				maxWeight = behaviorWeights[i];
		}

		for (int i = 0; i < behaviors.size(); i++) {
			Behavior b = behaviors.get(i);
			for (int j = 0; j < actions.size(); j++) {
				float weightedVote = b.getVote()[j] * actionHandler.getSituationWeightManager().getSituationWeights()[i] * actionHandler.getBehaviorWeights()[i] / maxWeight;
				float weightedVoteAlpha = weightedVote == 0 ? 0 : 1;
				g.setStroke(new BasicStroke(Math.abs(weightedVote * 10)));
				if (weightedVote >= 0) {
					g.setColor(new Color(0, 0, 0, weightedVoteAlpha));
				} else {
					g.setColor(new Color(1, 0, 0, weightedVoteAlpha));
				}
				g.drawLine(colsStartX[2] + boxW, i * (boxH + spaceH) + boxH / 2, colsStartX[4], j * (boxH + spaceH) + boxH / 2);
			}
		}
		g.setStroke(Util.thinLine);

		// Behaviors
		for (int i = 0; i < behaviors.size(); i++) {
			g.setColor(Color.BLACK);
			g.drawString(Util.fixedDecimals(actionHandler.getSituationWeightManager().getSituationWeights()[i], decimalPlaces), colsStartX[0], i * (boxH + spaceH) + stringHeight);
			g.drawString(Util.fixedDecimals(actionHandler.getBehaviorWeights()[i], decimalPlaces) + "", colsStartX[1], i * (boxH + spaceH) + stringHeight);

			g.setColor(Color.WHITE);
			g.fillRect(colsStartX[2], i * (boxH + spaceH), boxW, boxH);
			g.setColor(Color.BLUE);
			g.drawRect(colsStartX[2], i * (boxH + spaceH), boxW, boxH);
			g.setColor(Color.BLACK);
			g.drawString(behaviors.get(i).getName(), colsStartX[2] + 5, i * (boxH + spaceH) + stringHeight);
		}

		Color green = new Color(0, 0.8f, 0);
		// Actions
		for (int i = 0; i < actions.size(); i++) {
			g.setColor(Color.WHITE);
			g.fillRect(colsStartX[4], i * (boxH + spaceH), boxW, boxH);
			g.setColor(green);
			g.drawRect(colsStartX[4], i * (boxH + spaceH), boxW, boxH);

			g.setColor(Color.BLACK);
			g.drawString(actions.get(i).getName(), colsStartX[4] + 5, i * (boxH + spaceH) + stringHeight);
			String votes = Util.fixedDecimals(actionHandler.getVotes()[i], decimalPlaces);
			if (BestIndex == i) {
				g.setColor(green);
			}
			g.drawString(votes, colsStartX[5], i * (boxH + spaceH) + stringHeight);
		}
		g.translate(0, -headerSpace);

		// Draw velocity
		// g.translate(0, robot.getWorld().getHeight() * Gui.scale * 0.81f);
		// g.setColor(Color.BLACK);
		// g.drawString("Velocity: " + Util.fixedDecimals(robot.getBody().getLinearVelocity().length(), 2), 0, 0);
		// g.translate(0, -(robot.getWorld().getHeight() * Gui.scale * 0.81f));

		 g.translate(0, robot.getWorld().getHeight() * Gui.scale * 0.81f);
		 g.setColor(Color.BLACK);
		 g.drawString(phenotypeValues[9] + "", 0, 0);
		 g.drawString(phenotypeValues[10] + "", 100, 0);
		 g.drawString(phenotypeValues[11] + "", 200, 0);
		 g.drawString(phenotypeValues[12] + "", 300, 0);
		 g.translate(0, -(robot.getWorld().getHeight() * Gui.scale * 0.81f));

		// Draw sensor values
		g.setColor(Color.BLACK);
		g.translate(0, robot.getWorld().getHeight() * Gui.scale * 0.87f);
		ArrayList<LightSensor> sensors = robot.getSensorHandler().getLightSensors();

		int diffBetweenDistAndLight = 200;
		g.setFont(new Font("Arial", Font.BOLD, fontSize));
		g.drawString("Distance values", 0, 0);
		g.drawString("Light values", diffBetweenDistAndLight, 0);
		g.drawString("Holes detected", 2 * diffBetweenDistAndLight, 0);

		g.translate(0, stringHeight);
		g.setFont(new Font("Arial", Font.PLAIN, fontSize));
		int[] xPos = new int[] { 50, 75, 100, 50, 0, 25 };
		int[] yPos = new int[] { 0, boxH / 2, boxH, 2 * boxH, boxH, boxH / 2 };

		try {
			for (int i = 0; i < sensors.size(); i++) {
				g.setColor(Color.BLACK);
				g.drawString(Util.floatToDecimalAndExponentialFormat(sensors.get(i).getDistanceValue()), xPos[i], yPos[i]);

				g.setColor(sensors.get(i).getClosestColorToSensorColor());
				g.drawString(Util.floatToDecimalAndExponentialFormat(sensors.get(i).getLightValue()), xPos[i] + diffBetweenDistAndLight, yPos[i]);
				g.setColor(Color.BLACK);

				g.drawString("" + sensors.get(i).getHoleDetectionOutputString(), xPos[i] + 2 * diffBetweenDistAndLight, yPos[i]);
			}
		} catch (Exception e) {
		}

		g.translate(0, -(robot.getWorld().getHeight() * Gui.scale * 0.87f + stringHeight));
	}

	public ActionHandler getActionHandler() {
		return actionHandler;
	}

	@Override
	public void setup(Robot robot) {
		this.robot = robot;
		actionHandler.getSituationWeightManager().setRobot(robot);

	}

	public void addConnectionToPhenotype() {
		phenotype.addNumberOfConnections((robot.hasBeenInvolvedInConnection() ? 1 : 0));
	}

	public Phenotype<Float> getPhenotype() {
		return phenotype;
	}

	public Float[] getPhenotypeValues() {
		return phenotypeValues;
	}

}
