package managers;

import java.awt.Color;
import java.util.ArrayList;

import controller.ArbitrationController;

import options.SimulatorOptions;
import robot.LightSensor;
import robot.Robot;
import robot.SensorHandler;
import behaviors.AssembleRecoveryBehavior;
import behaviors.Behavior;
import behaviors.ConvergingBehavior;

public class SituationWeightManager {

	private float[] situationWeights;
	private ArrayList<Behavior> behaviors;

	private final int timeStartStagnation = 3;
	private final int timeIncreasedStagnation = 6;
	private boolean closeToRed;
	private ArbitrationController controller;
	private Robot robot;

	public SituationWeightManager(ArrayList<Behavior> behaviors) {
		situationWeights = new float[behaviors.size()];
		this.behaviors = behaviors;
	}

	// calculate new situation weights
	public void updateSituationWeights() {
		// Update variable used by multiple methods
		updateCloseToRed();
		
		// calc situation for CollisionAvoidance
		situationWeights[ActionHandler.COLLISION_AVOIDANCE] = calcCollisionAvoidanceSituation(behaviors.get(ActionHandler.COLLISION_AVOIDANCE));

		// calc situation for GoalConverging
		situationWeights[ActionHandler.GOAL_CONVERGING] = calcGoalConvergingSituation(behaviors.get(ActionHandler.GOAL_CONVERGING));

		// calc situation for Wander
		situationWeights[ActionHandler.WANDER] = 0.5f;

		// calc situation for Assemble
		situationWeights[ActionHandler.ASSEMBLE] = calcAssembleSituation(behaviors.get(ActionHandler.ASSEMBLE));

		// calc situation for AssembleStagnationRecovery
		situationWeights[ActionHandler.ASSEMBLE_RECOVERY] = calcAssembleStagnationRecoverySituation((AssembleRecoveryBehavior) behaviors.get(ActionHandler.ASSEMBLE_RECOVERY));

		// calc situation for StagnationRecovery
		situationWeights[ActionHandler.STAGNATION_RECOVERY] = calcStagnationRecoverySituation(behaviors.get(ActionHandler.STAGNATION_RECOVERY));

		situationWeights[ActionHandler.MAINTAIN_TURN] = 0.5f;

		situationWeights[ActionHandler.FOLLOW_LEADER] = calcFollowLeaderSituation();
		
		situationWeights[ActionHandler.HOLE_AVOIDANCE] = calcHoleAvoidanceSituation();

	}
	
	public void updateCloseToRed(){
		for (LightSensor s : robot.getSensorHandler().getLightSensors()) {
			if (s.getClosestColorToSensorColor() == Color.RED && s.getLightValue() > controller.lightThresholdClose) {
				closeToRed = true;
				return; 
			}
		}
		closeToRed = false;
	}

	public float[] getSituationWeights() {
		return situationWeights;
	}

	private float calcCollisionAvoidanceSituation(Behavior behavior) {
		if (robot.isConnected())
			return 0;

		float lowestDistance = controller.distanceThresholdFar;
		ArrayList<LightSensor> sensors = robot.getSensorHandler().getLightSensors();

		for (int sensorIndex = 0; sensorIndex < sensors.size(); sensorIndex++) {
			float distance = behavior.getDistanceFromSensor(sensorIndex);
			// Distance factor
			if (distance < lowestDistance) {
				Color closestColor = sensors.get(sensorIndex).getClosestColorToSensorColor();
				boolean isCloseToColorSignal = (closestColor == Color.GREEN || closestColor == Color.BLUE || closestColor == Color.YELLOW || closestColor == Color.RED);
				if (!(isCloseToColorSignal && behavior.isCloseToLight(sensorIndex))) {
					lowestDistance = distance;
				}
			}
		}

		if(lowestDistance < controller.distanceThresholdClose){
			return 1;
		}else if(lowestDistance < controller.distanceThresholdFar){
			return 0.5f;
		}else{
			return 0;
		}
	}

	private float calcGoalConvergingSituation(Behavior behavior) {
		float strongest = 0;
		ArrayList<LightSensor> sensors = robot.getSensorHandler().getLightSensors();

		for (int i = 0; i < sensors.size(); i++) {
			if (sensors.get(i).getClosestColorToSensorColor() == Color.RED) {
				float light = behavior.getLightFromSensor(i);
				if (light > strongest) {
					strongest = light;
				}
			}
		}
		
		if(strongest > controller.lightThresholdClose){
			return 1;
		}else if(strongest > controller.lightThresholdFar){
			return 0.5f;
		}else{
			return 0;
		}
	}

	private float calcAssembleSituation(Behavior behavior) {
		if(!SimulatorOptions.allowSelfAssembling)
			return 0;
		
		float strongest = 0f;

		ArrayList<LightSensor> sensors = robot.getSensorHandler().getLightSensors();

		for (int i = 0; i < sensors.size(); i++) {
			Color closestSensorColor = sensors.get(i).getClosestColorToSensorColor();
			if (((ConvergingBehavior) behavior).isRightColor(closestSensorColor) || (i == SensorHandler.BACK && closestSensorColor == Color.BLUE)) {
				float light = behavior.getLightFromSensor(i);
				if (light > strongest) {
					strongest = light;
				}
			}
		}

		if(strongest > controller.lightThresholdClose){
			return 1;
		}else if(strongest > controller.lightThresholdFar){
			return 0.5f;
		}else{
			return 0;
		}
	}

	private float calcAssembleStagnationRecoverySituation(AssembleRecoveryBehavior behavior) {
		if(behavior.isInRecoveryMode()){
			return 1;
		}
		if(!SimulatorOptions.allowSelfAssembling || closeToRed || !robot.isConnected())
			return 0;
		
		if (robot.getSensorHandler().getAccelerometer().getTimeSinceMoved() > timeIncreasedStagnation) {
			return 1;
		}else if (robot.getSensorHandler().getAccelerometer().getTimeSinceMoved() > timeStartStagnation) {
			return 0.5f;
		}else{
			return 0;
		}
	}

	private float calcStagnationRecoverySituation(Behavior behavior) {
		// returns 0 if it does have any connections or is close to red
		if (robot.isConnected() || closeToRed) {
			return 0f;
		}
		
		if (robot.getSensorHandler().getAccelerometer().getTimeSinceMoved() > timeIncreasedStagnation) {
			return 1;
		}else if (robot.getSensorHandler().getAccelerometer().getTimeSinceMoved() > timeStartStagnation) {
			return 0.5f;
		}else{
			return 0;
		}
	}

	private float calcFollowLeaderSituation() {
		return robot.isConnected() ? 0.5f : 0f; 
	}
	
	private float calcHoleAvoidanceSituation(){
		float situation = 0f;
		
		ArrayList<LightSensor> sensors = robot.getSensorHandler().getLightSensors();
		for(LightSensor s:sensors){
			if(s.isHoleDetected(LightSensor.HOLE_SENSOR_CLOSE)){
				return 1f;
			}
			else if(s.isHoleDetected(LightSensor.HOLE_SENSOR_FAR)){
				situation = 0.5f;
			}
		}
		return situation;
	}

	public void setRobot(Robot robot) {
		this.robot = robot;
		controller = (ArbitrationController) robot.getController();
	}
}
