package swarm;


import general.ChildPopulationMaker;
import general.Crossover;
import general.Developmental;
import general.EvaluatePopulationHandler;
import general.Fitness;
import general.Genotype;
import general.Individual;
import general.Mutation;
import general.PlotFitnessInEvolution;
import general.Population;
import general.Reproduction;
import general.SavePhenotype;

import java.util.ArrayList;

import options.EAOptions;
import options.SimulatorOptions;
import selection.adult.AdultSelection;
import selection.parent.ParentSelection;
import util.FileHandler;
import util.PerformanceTester;
import util.Util;

public class RunEA {

	/** Variables */
	private int nmbOfGenerations, populationSize, genLength, numberOfElites;
	private double crossoverRate, mutationRateOnIndividual;
	private AdultSelection adultSelection;
	private ParentSelection parentSelection;
	private EvaluatePopulationHandler evaluatePopulationHandler;
	private Crossover crossover;
	private Developmental developmentalType;
	private Mutation mutationType;
	private ChildPopulationMaker childPopulationMaker;
	private Fitness fitness;
	private float[][] statistics;

	/** Private Methods */

	private void extractOptions(EAOptions options) {
		nmbOfGenerations = options.nmbOfGenerations;
		populationSize = options.populationSize;
		genLength = options.genotypeLength;
		numberOfElites = options.numberOfElites;
		crossoverRate = options.crossoverRate;
		mutationRateOnIndividual = options.mutationRateOnIndividual;
		adultSelection = options.adultSelection;
		parentSelection = options.parentSelection;
		evaluatePopulationHandler = options.evaluatePopulationHandler;
		crossover = options.crossover;
		developmentalType = options.developmentalType;
		mutationType = options.mutationType;
		childPopulationMaker = options.childPopulationMaker;
		fitness = options.fitness;
		statistics = new float[nmbOfGenerations][3];
	}
	
	/** Public Methods */
	
	public Individual runEA(EAOptions options) {
		extractOptions(options);

		PerformanceTester.startTest("EA");
		// Setup and clear old
		Population.adultPopulation.clear();
		Population.parentPopulation.clear();
		Reproduction reproduction = new Reproduction();
		Individual bestFitnessIndividual = null;
		evaluatePopulationHandler.setFitness(fitness);
		PlotFitnessInEvolution plotter = new PlotFitnessInEvolution(nmbOfGenerations, options);

		// Initialize child population with random legal genotype
		childPopulationMaker.createRandomChildPopulation(populationSize, genLength);

		// Run each generation
		for (int gen = 0; gen < nmbOfGenerations; gen++) {
			if (options.measureTimeEveryGeneration) {
				PerformanceTester.startTest("generation " + gen);
			}

			// Development: Generate Phenotypes from Genotypes
			for (Individual child : Population.childPopulation) {
				child.setPhenotype(developmentalType.translate(child.getGenotype()));
			}

			// Test Fitness of Phenotypes
			evaluatePopulationHandler.doFitnessTest(Population.childPopulation);

			// Adult Selection
			adultSelection.select();
			
			// Save elites
			Population.makeElitePopulation(numberOfElites);

			// Parent Selection
			parentSelection.select();

			// Find best individual
			Individual bestOfGeneration = null;
			for (Individual adult : Population.adultPopulation) {
				if (bestOfGeneration == null || adult.getFitness() > bestOfGeneration.getFitness()) {
					bestOfGeneration = adult;
				}
			}

			if (bestFitnessIndividual == null || bestOfGeneration.getFitness() > bestFitnessIndividual.getFitness()) {
				bestFitnessIndividual = bestOfGeneration;
				if (options.savePhenotypes == SavePhenotype.SAVE_DURING)
					FileHandler.savePhenoype(bestFitnessIndividual, fitness.getTestName());
			}

			// Plott fitness graph
			plotter.updatePlot(gen);
			if(options.saveStatistics){
				statistics[gen] = getGenerationStatistics();
			}

			// Reproduction
			ArrayList<Genotype<?>> genotypes = reproduction.reproduce(crossoverRate, mutationRateOnIndividual, crossover, mutationType, numberOfElites);
			childPopulationMaker.makeChildrenFromGenotypes(genotypes, gen + 1);

			if (options.measureTimeEveryGeneration) {
				PerformanceTester.endTest("generation " + gen);
			}
		}

		// Printing
		PerformanceTester.endTest("EA");
		System.out.println("Best " + bestFitnessIndividual.getFitness());
		
		// Save best phenotype and plot
		if (options.savePhenotypes == SavePhenotype.SAVE_WHEN_FINISHED) {
			FileHandler.savePhenoype(bestFitnessIndividual, fitness.getTestName());
		}
		plotter.writePlotImage(fitness.getTestName());
		
		
		return bestFitnessIndividual;
	}
	
	private float[] getGenerationStatistics(){
		
		float sum = 0.0f;
		float max = Float.MIN_VALUE;

		for (Individual adult : Population.adultPopulation) {
			sum += adult.getFitness();
			if (adult.getFitness() > max)
				max = (float) adult.getFitness();
		}
		float average = sum / Population.adultPopulation.size();

		float[] list = new float[Population.adultPopulation.size()];
		for (int i = 0; i < list.length; i++) {
			list[i] = (float) Population.adultPopulation.get(i).getFitness();
		}

		float standardDeviation = Util.standardDeviation(list, average);
		
		return new float[]{average, max, standardDeviation};
	}

	public float[][] getStatistics() {
		return statistics;
	}

}
