package environment_objects;

import general.Gui;

import java.awt.Color;
import java.awt.Graphics2D;
import java.util.ArrayList;

import objects.BodyObject;
import objects.Light;

import org.jbox2d.collision.shapes.PolygonShape;
import org.jbox2d.common.Vec2;
import org.jbox2d.dynamics.BodyType;
import org.jbox2d.dynamics.FixtureDef;

import util.Tuple;
import util.Util;

public class Ground extends BodyObject {
	
	/**
	 * Constants
	 */

	private final float normalFrictionKinetic = 0.4f;
	private final float normalFrictionStatic = 0.5f;
	private final float iceFrictionKinetic = 0.03f;//0.02f; // http://no.wikipedia.org/wiki/Friksjon
	private final float iceFrictionStatic = 0.04f;//0.03f;
	private final Color groundColor = new Color(0.9f, 0.9f, 0.9f);
	private final Color startAreaColor = new Color(0.2f, 0.2f, 0.2f, 0.2f);
	
	/**
	 * Variable
	 */
	
	private ArrayList<Ice> iceAreas;
	private ArrayList<WindArea> windAreas;
	private ArrayList<Hole> holeAreas;
	private float[] startingArea = null;
	private final float width, height;

	/**
	 * Constructors
	 */

	public Ground(Vec2 position, float width, float height) {
		super(position, BodyType.STATIC);
		shape = new PolygonShape();
		((PolygonShape) shape).setAsBox(width, height);
		this.width = width;
		this.height = height;
		iceAreas = new ArrayList<Ice>();
		windAreas = new ArrayList<WindArea>();
		holeAreas = new ArrayList<Hole>();
	}

	@Override
	public FixtureDef setFixtureDef() {
		FixtureDef fd = new FixtureDef();
		fd.shape = shape;
		fd.isSensor = true;
		fd.friction = normalFrictionKinetic;
		return fd;
	}

	@Override
	public void update() {
	}

	public float getSurfaceFriction(Vec2 position, boolean isMoving) {
		for (Ice ice : iceAreas) {
			if (ice.isOver(position)) {
				if(isMoving)
					return iceFrictionKinetic;
				else
					return iceFrictionStatic;
			}
		}
		if(isMoving)
			return normalFrictionKinetic;
		else
			return normalFrictionStatic;
	}
	
	public Vec2 getCloseHole(Vec2 position) {
		for (Hole h : holeAreas) {
			if(h.isOver(position))
				return h.getPosition();
		}
		return null;
	}
	
	public boolean isOverHole(Vec2 position) {
		for (Hole h : holeAreas) {
			if(h.isOver(position))
				return true;
		}
		return false;
	}

	public WindArea getWindArea(Vec2 position) {
		for (WindArea windArea : windAreas) {
			if (windArea.isOver(position)) {
				return windArea;
			}
		}
		return null;
	}

	public void addWindArea(WindArea windArea) {
		windAreas.add(windArea);
	}

	public void addIceArea(Ice iceArea) {
		iceAreas.add(iceArea);
	}

	public void setStartingArea(float x, float y, float dx, float dy) {
		this.startingArea = new float[] { x, y, dx, dy };
	}

	@Override
	public void draw(Graphics2D g) {
		int x = (int) (getBody().getPosition().x * Gui.scale);
		int y = (int) (getBody().getPosition().y * Gui.scale);
		int dX = (int) (width * Gui.scale);
		int dY = (int) (height * Gui.scale);
		g.translate(x, y);
		g.setColor(groundColor);

		g.fillRect(0, 0, dX, dY);

		g.translate(-x, -y);

		if (startingArea != null) {
			g.setStroke(Util.thickLine);
			g.setColor(startAreaColor);
			g.fillRect((int) (startingArea[0] * Gui.scale), (int) (startingArea[1] * Gui.scale), (int) (startingArea[2] * Gui.scale), (int) (startingArea[3] * Gui.scale));
			g.setStroke(Util.dashedLine);
			g.setColor(Color.YELLOW);
			g.drawRect((int) (startingArea[0] * Gui.scale), (int) (startingArea[1] * Gui.scale), (int) (startingArea[2] * Gui.scale), (int) (startingArea[3] * Gui.scale));
			g.drawString("Start area", (int) (startingArea[0] * Gui.scale + 5), (int) (startingArea[1] * Gui.scale + 15));
			g.setStroke(Util.thinLine);
		}

		for (Ice ice : iceAreas) {
			ice.draw(g);
		}
		for (Hole h : holeAreas) {
			h.draw(g);
		}
		for (WindArea windArea : windAreas) {
			windArea.draw(g);
		}
	}

	public void drawOver(Graphics2D g) {
		for (WindArea windArea : windAreas) {
			windArea.drawOver(g);
		}
	}
	
	public float[] getStartingArea() {
		return startingArea;
	}

	@Override
	public Tuple<Light, Float> getLightProperties(float normalAngle) {
		return null;
	}

	public void addHole(Hole hole) {
		holeAreas.add(hole);
	}
	
	public ArrayList<Hole> getHoleAreas(){
		return holeAreas;
	}

	@Override
	public boolean isOccupied(Vec2 position) {
		return false;
	}
	
	public void updateWindAreas(){
		for (WindArea windArea : windAreas) {
			windArea.updateWindFactors();
		}
	}
}
