package environment_objects;

import general.Gui;

import java.awt.Color;
import java.awt.Graphics2D;

import objects.BodyObject;
import objects.LightSource;
import options.SimulatorOptions;

import org.jbox2d.common.Vec2;

import robot.Robot;
import util.Util;

public class WindArea {
	public static final float windConstants = 0.5f * 1.29f * 0.8f * Robot.size * 2 * Robot.height;// 0.5f * ro * C * A;
	public static final int EAST = 0, SOUTH = 90, WEST = 180;
	private int windDirection;
	private Vec2 position;
	private Vec2 size;
	private Vec2 windForce;
	private float windVelocity;
	private float[][] windFactors;
	private boolean[][] windUpdated;
	private int gridRows;
	private int gridCols;
	private World world;
	private final float windIncrease = 1f;
	public static boolean drawOutline = false;

	public WindArea(Vec2 position, Vec2 size, float windVelocity, int windDirection, World world) {
		this.position = position;
		this.size = size;
		this.windVelocity = windVelocity;
		this.world = world;
		this.windDirection = windDirection;
		calculateWindForce();
		gridCols = (int) (size.x / SimulatorOptions.windCellSize + 1);
		gridRows = (int) (size.y / SimulatorOptions.windCellSize + 1);
		windFactors = new float[gridRows][gridCols];
		windUpdated = new boolean[gridRows][gridCols];
	}

	public void setObjectAt(Vec2 pos, float radius) {
		radius = (float) (radius * 0.9);
		float radiusSquared = radius * radius;

		float inc = 0.01f;
		float xStart = -radius, yStart = -radius, endX = radius, endY = radius;
		switch (windDirection) {
		case EAST:
			xStart = 0;
			break;
		case SOUTH:
			yStart = 0;
			break;
		case WEST:
			endX = 0;
			break;
		}

		Vec2 posOnGrid = pos.sub(position);
		for (float x = xStart; x <= endX; x += inc) {
			for (float y = yStart; y <= endY; y += inc) {
				Vec2 temp = new Vec2(x, y);
				float lengthSquared = temp.lengthSquared();
				if (radiusSquared > lengthSquared) {
					int xIndex = Math.round((x - SimulatorOptions.windCellSize / 2 + posOnGrid.x) / SimulatorOptions.windCellSize);
					int yIndex = Math.round((y - SimulatorOptions.windCellSize / 2 + posOnGrid.y) / SimulatorOptions.windCellSize);
					if (xIndex < 0 || xIndex > gridCols - 1)
						continue;
					if (yIndex < 0 || yIndex > gridRows - 1)
						continue;
					windFactors[yIndex][xIndex] = 0f;
					windUpdated[yIndex][xIndex] = true;
				}
			}
		}
	}

	private float windIncreasePerCell() {
		return windIncrease * SimulatorOptions.windCellSize;
	}

	public void updateWindFactors() {
		// reset
		reset();

		// remove ones with robot
		for (BodyObject b : world.getWorldObjects()) {
			if (b instanceof Robot) {
				setObjectAt(b.getBody().getPosition(), Robot.size);
			} else if (b instanceof LightSource) {
				setObjectAt(b.getBody().getPosition(), ((LightSource) b).getSize());
			}
		}

		// update rest sequentially
		updateSequentially();
	}

	private void updateSequentially() {
		switch (windDirection) {
		case EAST:
			for (int x = 1; x < gridCols; x++) {
				for (int y = 0; y < gridRows; y++) {
					if (!windUpdated[y][x]) {
						if (y == 0) {
							windFactors[y][x] = Math.min(1f, (float) (0.3 * windFactors[y + 1][x - 1] + 0.7 * windFactors[y][x - 1] + windIncreasePerCell()));
						} else if (y == gridRows - 1) {
							windFactors[y][x] = Math.min(1f, (float) (0.3 * windFactors[y - 1][x - 1] + 0.7 * windFactors[y][x - 1] + windIncreasePerCell()));
						} else {
							windFactors[y][x] = Math.min(1f, (float) (0.3 * windFactors[y - 1][x - 1] + 0.4 * windFactors[y][x - 1] + 0.3 * windFactors[y + 1][x - 1] + windIncreasePerCell()));
						}
					}
				}
			}
			break;
		case SOUTH:
			for (int y = 1; y < gridRows; y++) {
				for (int x = 0; x < gridCols; x++) {
					if (!windUpdated[y][x]) {
						if (x == 0) {
							windFactors[y][x] = Math.min(1f, (float) (0.7 * windFactors[y - 1][x] + 0.3 * windFactors[y - 1][x + 1] + windIncreasePerCell()));
						} else if (x == gridCols - 1) {
							windFactors[y][x] = Math.min(1f, (float) (0.7 * windFactors[y - 1][x] + 0.3 * windFactors[y - 1][x - 1] + windIncreasePerCell()));
						} else {
							windFactors[y][x] = Math.min(1f, (float) (0.3 * windFactors[y - 1][x - 1] + 0.4 * windFactors[y - 1][x] + 0.3 * windFactors[y - 1][x + 1] + windIncreasePerCell()));
						}
					}
				}
			}
			break;
		case WEST:
			for (int x = gridCols - 2; x >= 0; x--) {
				for (int y = 0; y < gridRows; y++) {
					if (!windUpdated[y][x]) {
						if (y == 0) {
							windFactors[y][x] = Math.min(1f, (float) (0.7 * windFactors[y][x + 1] + 0.3 * windFactors[y + 1][x + 1] + windIncreasePerCell()));
						} else if (y == gridRows - 1) {
							windFactors[y][x] = Math.min(1f, (float) (0.7 * windFactors[y][x + 1] + 0.3 * windFactors[y - 1][x + 1] + windIncreasePerCell()));
						} else {
							windFactors[y][x] = Math.min(1f, (float) (0.3 * windFactors[y - 1][x + 1] + 0.4 * windFactors[y][x + 1] + 0.3 * windFactors[y + 1][x + 1] + windIncreasePerCell()));
						}
					}
				}
			}
			break;
		}

	}

	private void reset() {
		switch (windDirection) {
		case EAST:
			for (int y = 0; y < gridRows; y++) {
				windFactors[y][0] = 1.0f;

			}
			for (int x = 1; x < gridCols; x++) {
				for (int y = 0; y < gridRows; y++) {
					windUpdated[y][x] = false;
				}
			}
			break;
		case SOUTH:
			for (int x = 0; x < gridCols; x++) {
				windFactors[0][x] = 1.0f;

			}
			for (int y = 1; y < gridRows; y++) {
				for (int x = 0; x < gridCols; x++) {
					windUpdated[y][x] = false;
				}
			}
			break;
		case WEST:
			for (int y = 0; y < gridRows; y++) {
				windFactors[y][gridCols - 1] = 1.0f;
			}
			for (int x = 0; x < gridCols - 1; x++) {
				for (int y = 0; y < gridRows; y++) {
					windUpdated[y][x] = false;
				}
			}
			break;
		}

	}

	private void calculateWindForce() {
		// This is done here once for performance reasons.
		// If more objects than robots can be affected by the wind then this must be calculated for the individual as C and A is dependent on object
		windForce = new Vec2();
		if (windVelocity < 0) {
			System.err.println("negative wind is not supported, change direction instead");
			System.exit(0);
		}

		switch (windDirection) {
		case EAST:
			windForce.x = (float) (windConstants * windVelocity * windVelocity);
			break;
		case SOUTH:
			windForce.y = (float) (windConstants * windVelocity * windVelocity);
			break;
		case WEST:
			windForce.x = (float) (-windConstants * windVelocity * windVelocity);
			break;
		default:
			System.err.println(windDirection + " is not a known direction");
			System.exit(0);
		}
	}

	public float getWindHitRateFactor(Vec2 pos, float radius) {
		Vec2 posOnGrid = pos.sub(position);
		double[] samples = new double[(int) (radius * 2 / SimulatorOptions.windCellSize)];
		try {
			switch (windDirection) {
			case EAST:
				for (int i = 0; i < samples.length; i++) {
					int x = Math.round((posOnGrid.x - radius) / SimulatorOptions.windCellSize);
					int y = Math.round((posOnGrid.y + i * 2 * radius / samples.length - radius) / SimulatorOptions.windCellSize);
					samples[i] = windFactors[y][x];
				}
				break;
			case SOUTH:
				for (int i = 0; i < samples.length; i++) {
					int x = Math.round((posOnGrid.x + i * 2 * radius / samples.length - radius) / SimulatorOptions.windCellSize);
					int y = Math.round((posOnGrid.y - radius) / SimulatorOptions.windCellSize);
					samples[i] = windFactors[y][x];
				}
				break;
			case WEST:
				for (int i = 0; i < samples.length; i++) {
					int x = Math.round((posOnGrid.x + radius) / SimulatorOptions.windCellSize);
					int y = Math.round((posOnGrid.y + i * 2 * radius / samples.length - radius) / SimulatorOptions.windCellSize);
					samples[i] = windFactors[y][x];
				}
				break;
			}
		} catch (Exception e) {
			return 1;
		}

		return (float) Util.average(samples);
	}

	public boolean isOver(Vec2 positionBody) {
		if (positionBody.x < position.x)
			return false;
		if (positionBody.y < position.y)
			return false;
		if (positionBody.y > position.y + size.y)
			return false;
		if (positionBody.x > position.x + size.x)
			return false;
		return true;
	}

	public Vec2 getWindForce() {
		return windForce;
	}

	public void draw(Graphics2D g) {
		if (!SimulatorOptions.drawWindReduction ||SimulatorOptions.onlyShowEnvironment)
			return;
		int guiX = (int) (position.x * Gui.scale);
		int guiY = (int) (position.y * Gui.scale);
		// Area
		g.translate(guiX, guiY);

		g.setColor(Color.BLACK);
		for (int y = 0; y < gridRows; y++) {
			for (int x = 0; x < gridCols; x++) {
				float alpha;
				if (1 - windFactors[y][x] > 0.1) {
					alpha = 1;
				} else {
					alpha = 1 - windFactors[y][x];
				}
				g.setColor(new Color(windFactors[y][x], windFactors[y][x], windFactors[y][x], alpha));
				g.fillRect((int) (SimulatorOptions.windCellSize * x * Gui.scale), (int) (SimulatorOptions.windCellSize * y * Gui.scale), (int) Math.ceil(SimulatorOptions.windCellSize * Gui.scale), (int) Math.ceil(SimulatorOptions.windCellSize * Gui.scale));

				// String w = Util.fixedDecimals(windFactors[y][x], 1);
				// g.setColor(Color.BLACK);
				// g.drawString(w, cellSize * (x + 0.5f) * Gui.scale - Util.getStringWithOnScreen(g, w) / 2, cellSize * (y + 0.5f) * Gui.scale);
			}
		}

		g.translate(-guiX, -guiY);
	}

	public void drawOver(Graphics2D g) {
		int guiX = (int) (position.x * Gui.scale);
		int guiY = (int) (position.y * Gui.scale);
		int dX = (int) (size.x * Gui.scale);
		int dY = (int) (size.y * Gui.scale);
		// Area
		g.translate(guiX, guiY);
		if (drawOutline) {
			g.setColor(new Color(0f, 1f, 0f, 0.1f));
			g.fillRect(0, 0, dX, dY);
			g.setColor(new Color(0f, 1f, 0f, 0.5f));
			g.setStroke(Util.dashedLine);
			g.drawRect(0, 0, dX, dY);
		}
		Vec2 middle = new Vec2(dX / 2, dY / 2);
		g.translate(middle.x, middle.y);
		g.setStroke(Util.thinLine);
		g.setColor(Color.BLACK);

		String s = "Wind: " + Math.round(windVelocity) + " m/s";
		g.drawString(s, -Util.getStringWithOnScreen(g, s) / 2, -50);

		drawArrow(g, 0);
		if (!drawOutline) {
			drawArrow(g, Gui.scale / 10);
			drawArrow(g, -Gui.scale / 10);
		}
		g.translate(-middle.x, -middle.y);

		g.translate(-guiX, -guiY);
	}

	public void drawArrow(Graphics2D g, float delta) {
		if(windForce.length() == 0){
			return;
		}
		// Arrow
		float angle = Util.getVec2Angle(windForce);
		Vec2 length = Util.createVec2(Gui.scale / 10, angle);
		float x = (float) (-delta * Math.sin(angle));
		float y = (float) (delta * Math.cos(angle));

		// float x = 0, y= 0;
		Vec2 arrowStart = (new Vec2(x, y)).sub(length.mul(0.5f));
		Vec2 arrowPoint = arrowStart.add(length);

		g.setColor(Color.BLACK);
		g.drawLine(Math.round(arrowStart.x), Math.round(arrowStart.y), Math.round(arrowPoint.x), Math.round(arrowPoint.y));

		float downAngle = Util.normalizeAngle((float) (angle + Math.toRadians(150)));
		float upAngle = Util.normalizeAngle((float) (angle - Math.toRadians(150)));

		Vec2 down = (Util.createVec2(Gui.scale / 100, downAngle)).add(arrowPoint);
		Vec2 up = (Util.createVec2(Gui.scale / 100, upAngle)).add(arrowPoint);

		g.drawLine(Math.round(arrowPoint.x), Math.round(arrowPoint.y), Math.round(up.x), Math.round(up.y));
		g.drawLine(Math.round(arrowPoint.x), Math.round(arrowPoint.y), Math.round(down.x), Math.round(down.y));

	}
}
