package environment_objects;


import java.awt.Color;
import java.util.ArrayList;

import joints.AirResistanceJoint;
import joints.ConnectionJoint;
import joints.GroundFrictionJoint;
import objects.BodyObject;
import objects.LightSource;
import objects.Wall;

import org.jbox2d.common.Vec2;
import org.jbox2d.dynamics.joints.DistanceJoint;
import org.jbox2d.dynamics.joints.FrictionJoint;

import controller.ArbitrationController;

import robot.DeadRobot;
import robot.Robot;

public class World extends org.jbox2d.dynamics.World {
	private float width;
	private float height;
	private ArrayList<BodyObject> bodyObjects;
	private ArrayList<BodyObject> removeList;
	private ArrayList<Robot> robots;
	private ArrayList<DeadRobot> deadRobots;
	private float worldLightValue;
	private float[] worldLightColorIntensity;
	private BodyObject goalObject;

	private ArrayList<ConnectionJoint> connectionJoints;

	private Ground ground;

	public World(float width, float height) {
		super(new Vec2(0.0f, 0.0f));
		bodyObjects = new ArrayList<BodyObject>();
		robots = new ArrayList<Robot>();
		connectionJoints = new ArrayList<ConnectionJoint>();
		removeList = new ArrayList<BodyObject>();
		deadRobots = new ArrayList<DeadRobot>();

		this.width = width;
		this.height = height;
		worldLightValue = 0.0001f;
		worldLightColorIntensity = new float[] { worldLightValue, worldLightValue, worldLightValue };

		// Ground
		ground = new Ground(new Vec2(0.0f, 0.0f), width, height);
		addObject(ground);

		// Outer walls
		addObject(new Wall(new Vec2(0.0f, 0.0f), 0.0f, height));
		addObject(new Wall(new Vec2(0.0f, 0.0f), width, 0.0f));
		addObject(new Wall(new Vec2(width, 0.0f), 0.0f, height));
		addObject(new Wall(new Vec2(0.0f, height), width, 0.0f));
	}

	public float getWorldLightValue() {
		return worldLightValue;
	}

	public void setWorldLightValue(float worldLightValue) {
		this.worldLightValue = worldLightValue;
	}

	public float[] getWorldLightColorIntensity() {
		return worldLightColorIntensity;
	}

	public void addObject(BodyObject object) {
		bodyObjects.add(object);
		object.setBody(createBody(object.getBodyDef()));

		if (object instanceof Robot) {
			setFrictionJointsOnRobot((Robot) object);
			robots.add((Robot) object);
		}
	}

	public void addConnectionJoint(ConnectionJoint connectionJoint) {
		connectionJoint.setJoint((DistanceJoint) createJoint(connectionJoint.getJointDef()));
		connectionJoints.add(connectionJoint);
	}

	public void removeConnectionJoint(Robot a) {
		for (ConnectionJoint j : connectionJoints) {
			if (j.equals(a.getConnectionJoint())) {
				connectionJoints.remove(j);
				destroyJoint(j.getJoint());
				a.setConnectionJoint(null);
				return;
			}
		}
	}


	public void removeAllConnectionJointsWithRobot(Robot a) {
		removeConnectionJoint(a);
		ArrayList<ConnectionJoint> removeConnectionList = new ArrayList<ConnectionJoint>();
		for (ConnectionJoint j : connectionJoints) {
			if (j.getRobotB() == a) {
				removeConnectionList.add(j);
			}
		}
		while (removeConnectionList.size() > 0) {
			ConnectionJoint j = removeConnectionList.get(0);
			connectionJoints.remove(j);
			destroyJoint(j.getJoint());
			j.getRobotA().setConnectionJoint(null);
			j.getRobotA().getFrontLight().setColor(Color.BLUE);
			removeConnectionList.remove(j);
		}
	}

	public ArrayList<ConnectionJoint> getConnectionJoints() {
		return connectionJoints;
	}

	public boolean isJointBetween(Robot a, Robot b) {
		if ((a.getConnectionJoint() != null && a.getConnectionJoint().getRobotB().equals(b)) || (b.getConnectionJoint() != null && b.getConnectionJoint().getRobotB().equals(a))) {
			return true;
		}
		return false;
	}

	public void setFrictionJointsOnRobot(Robot robot) {
		GroundFrictionJoint groundFrictionJoint = new GroundFrictionJoint(robot, ground);
		FrictionJoint joint = (FrictionJoint) createJoint(groundFrictionJoint.getJointDef());
		groundFrictionJoint.setJoint(joint);
		robot.setFrictionJoint(groundFrictionJoint);

		AirResistanceJoint airResistanceJoint = new AirResistanceJoint(robot, ground);
		FrictionJoint airJoint = (FrictionJoint) createJoint(airResistanceJoint.getJointDef());
		airResistanceJoint.setJoint(airJoint);
		robot.setAirResistanceJoint(airResistanceJoint);
	}

	public ArrayList<BodyObject> getWorldObjects() {
		return bodyObjects;
	}

	public ArrayList<Robot> getRobots() {
		return robots;
	}

	public Ground getGround() {
		return ground;
	}

	public float getWidth() {
		return width;
	}

	public float getHeight() {
		return height;
	}

	public BodyObject getGoalObject() {
		return goalObject;
	}

	public void setGoalObject(BodyObject goalObject) {
		this.goalObject = goalObject;
		((LightSource) goalObject).setGoal(true);
	}

	public void removeObject(BodyObject bodyObject) {
		removeList.add(bodyObject);
	}
	
	public boolean isOccupied(Vec2 position){
		for (BodyObject b : bodyObjects) {
			if(b.isOccupied(position))
				return true;
		}
		return false;
	}

	public void clearRemoved() {
		while (removeList.size() > 0) {
			BodyObject bodyObject = removeList.get(0);

			// Remove from lists
			bodyObjects.remove(bodyObject);
			if (bodyObject instanceof Robot) {
				Robot r = (Robot) bodyObject;
				robots.remove(r);
				if(r.getController() instanceof ArbitrationController){
					if(((ArbitrationController)(r.getController())).getPhenotype() != null){
						((ArbitrationController)(r.getController())).addConnectionToPhenotype();
					}
				}
				
				removeAllConnectionJointsWithRobot(r);
				deadRobots.add(new DeadRobot(r.getBody().getPosition(), r.getBody().getAngle()));
			}
			destroyBody(bodyObject.getBody());
			// Remove connections
			removeList.remove(bodyObject);
		}
	}

	public ArrayList<DeadRobot> getDeadRobots() {
		return deadRobots;
	}
	
}
