package general;

import java.awt.Color;
import java.awt.Container;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.RenderingHints;
import java.awt.Toolkit;
import java.awt.event.KeyEvent;
import java.awt.event.KeyListener;
import java.awt.event.MouseEvent;
import java.awt.event.MouseListener;
import java.awt.image.BufferedImage;
import java.io.BufferedWriter;
import java.io.File;
import java.io.FileWriter;
import java.io.IOException;

import javax.imageio.ImageIO;
import javax.swing.JFrame;
import javax.swing.JPanel;

import joints.ConnectionJoint;

import options.SimulatorOptions;

import org.jbox2d.common.Vec2;

import controller.ArbitrationController;

import environment_objects.World;

import objects.BodyObject;
import robot.DeadRobot;
import robot.Robot;
import util.Util;

public class Gui extends JPanel implements MouseListener, KeyListener {
	private static final long serialVersionUID = 1L;

	public static boolean exit_on_close = true;
	private JFrame frame = null;
	private World world;
	public static int scale = 1;
	private final int borderPX = 5;
	private long frameTimeMS;
	public static float guiSpeedFactor = 1f;
	public static String setUpName;

	private int controllerDrawWidth = 628;
	private float elapsedTime;
	private double currentFitness = 0.0;
	private Robot drawRobotController;
	private float timeStep;
	private int lastImageSecond = -1;
	private int simWidth;
	private int simHeight;
	private String screenshotFolder;

	public Gui(World world, float timeStep) {
		this.world = world;
		if (SimulatorOptions.scaleGuiToScreen)
			scale = (int) (0.9 * (Toolkit.getDefaultToolkit().getScreenSize().getHeight() / world.getHeight() - 30 - 40));
		this.timeStep = timeStep;
		// Calculate frameTimeMS based on timestep and slowmotionFactor
		addMouseListener(this);
		addKeyListener(this);
		makeFrame();
		frameTimeMS = (long) (1000f * timeStep / guiSpeedFactor);
//		makeFolder();
	}

	private void makeFolder() {
		screenshotFolder = "C:/swarm_results/" + setUpName + " - " + Util.dateAsString() + " - " + System.nanoTime() + "/";
		File theDir = new File(screenshotFolder);

		// if the directory does not exist, create it
		if (!theDir.exists()) {
			theDir.mkdir();
		}
		try {
			ArbitrationController controller = (ArbitrationController)world.getRobots().get(0).getController();
			Float[] values = controller.getPhenotypeValues();
			String content = "";
			for (int i = 0; i < values.length; i++) {
				content += i + ":\r\t"+ values[i] + "\r\n"; 
			}
 
			File file = new File(screenshotFolder + "phenotype.txt");
 
			// if file doesnt exists, then create it
			if (!file.exists()) {
				file.createNewFile();
			}
 
			FileWriter fw = new FileWriter(file.getAbsoluteFile());
			BufferedWriter bw = new BufferedWriter(fw);
			bw.write(content);
			bw.close();
		} catch (IOException e) {
			e.printStackTrace();
		}
		
	}

	private void makeFrame() {
		frame = new JFrame("Simulator");
		frame.setLocation(400, 0);
		frame.setContentPane(this);
		frame.requestFocus();
		simWidth = (int) (scale * world.getWidth() + 8 + borderPX * 2);
		simHeight = (int) (scale * world.getHeight() + 30 + borderPX * 2);

		if (SimulatorOptions.drawController) {
			frame.setSize(simWidth + controllerDrawWidth, simHeight);
		} else {
			frame.setSize(simWidth, simHeight);
		}
		frame.setResizable(false);
		frame.setVisible(true);
	}

	public void exitGUI() {
		frame.dispose();
	}

	public void repaintGUI(float elapsedTime, double currentFitness) {

		if (exit_on_close) {
			frame.setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);
		} else {
			frame.setDefaultCloseOperation(JFrame.DISPOSE_ON_CLOSE);
		}
		this.elapsedTime = elapsedTime;
		this.currentFitness = currentFitness;
		if (SimulatorOptions.drawController) {
			simWidth = (int) (scale * world.getWidth() + 8 + borderPX * 2);
			simHeight = (int) (scale * world.getHeight() + 30 + borderPX * 2);
			frame.setSize(simWidth + controllerDrawWidth, simHeight);
		}

		repaint();

		if ((int) elapsedTime > lastImageSecond) {
			lastImageSecond = (int) elapsedTime;
			takeScreenShot();
		}

		try {
			Thread.sleep(frameTimeMS);
		} catch (InterruptedException e) {
			e.printStackTrace();
		}
	}

	protected void paintComponent(Graphics g) {
		super.paintComponent(g);
		requestFocusInWindow();
		Graphics2D g2 = (Graphics2D) g;
		g2.setRenderingHint(RenderingHints.KEY_ANTIALIASING, RenderingHints.VALUE_ANTIALIAS_ON);
		g2.setRenderingHint(RenderingHints.KEY_TEXT_ANTIALIASING, RenderingHints.VALUE_TEXT_ANTIALIAS_ON);
		g.translate(borderPX, borderPX);
		try {
			for (BodyObject object : world.getWorldObjects()) {
				object.draw((Graphics2D) g);
			}
			for (DeadRobot d : world.getDeadRobots()) {
				d.draw((Graphics2D) g);
			}
		} catch (Exception e) {
		}

		for (ConnectionJoint connectionJoint : world.getConnectionJoints()) {
			connectionJoint.draw((Graphics2D) g);
		}

		world.getGround().drawOver((Graphics2D) g);
		g.setColor(Color.BLACK);
		if (!SimulatorOptions.onlyShowEnvironment)
			g.drawString("Duration: " + Util.fixedDecimals(elapsedTime, 1), 5, 15);
		g.drawString("Width: " + world.getWidth() + " m", (int) (scale * world.getWidth() - 75), 15);
		g.drawString("Height: " + world.getHeight() + " m", 5, (int) (scale * world.getHeight() - 5));
		String fitness = Util.fixedDecimals((float) currentFitness, 2);
		if (!SimulatorOptions.onlyShowEnvironment)
			g.drawString("Fitness: " + fitness, (int) (borderPX + scale * world.getWidth() - 80), (int) (scale * world.getHeight() - 5));
		g.translate(-borderPX, -borderPX);

		// Clear outside walls
		g.setColor(Color.WHITE);
		g.fillRect(0, 0, (int) (world.getWidth() * scale + 20), borderPX);
		g.fillRect(0, 0, borderPX, (int) (world.getHeight() * scale + 40));
		g.fillRect((int) (world.getWidth() * scale + borderPX + 1), 0, (int) (world.getWidth() * scale + controllerDrawWidth + 2 * borderPX), (int) (world.getHeight() * scale + 40));
		g.fillRect(borderPX, (int) (world.getHeight() * scale + borderPX + 1), (int) (world.getWidth() * scale + 20), 40);
		g.setColor(Color.RED);
		g.drawRect(borderPX, borderPX, (int) (world.getWidth() * scale), (int) (world.getHeight() * scale));

		// Controller
		if (SimulatorOptions.drawController && world.getRobots().size() > 0) {
			if (drawRobotController == null) {
				drawRobotController = world.getRobots().get(0);
				drawRobotController.setDrawController(true);
			}
			g.translate((int) (world.getWidth() * Gui.scale + 30), borderPX);
			if (drawRobotController.getMotorIsRunning()) {
				drawRobotController.getController().draw((Graphics2D) g);
			} else {
				drawRobotController = null;
			}
		}
	}

	@Override
	public void mouseClicked(MouseEvent e) {
	}

	@Override
	public void mousePressed(MouseEvent e) {
	}

	@Override
	public void mouseReleased(MouseEvent e) {
		if (!SimulatorOptions.drawController)
			return;
		if (e.getButton() == MouseEvent.BUTTON1 && e.getX() < (scale * world.getWidth() + 8 + borderPX)) {
			for (Robot r : world.getRobots()) {
				Vec2 pos = r.getBody().getPosition().mul(scale);
				pos.x += borderPX;
				pos.y += borderPX;
				Vec2 mouse = new Vec2(e.getX(), e.getY());
				float length = mouse.sub(pos).length();
				if (length < Robot.getSize() * scale) {
					drawRobotController.setDrawController(false);
					drawRobotController = r;
					r.setDrawController(true);
					return;
				}
			}
		}
	}

	@Override
	public void mouseEntered(MouseEvent e) {
	}

	@Override
	public void mouseExited(MouseEvent e) {
	}

	@Override
	public void keyTyped(KeyEvent e) {
	}

	@Override
	public void keyPressed(KeyEvent e) {
		int integer = 0;
		try {
			integer = Integer.parseInt(String.valueOf(e.getKeyChar()));
		} catch (Exception exception) {
		}

		if (e.getKeyChar() == '+') {
			guiSpeedFactor *= 1.15f;
		} else if (e.getKeyChar() == '-') {
			guiSpeedFactor *= 0.9f;

			if (guiSpeedFactor < 0.03) {
				guiSpeedFactor = 0.03f;
			}
		} else if (e.getKeyChar() == '*') {
			guiSpeedFactor = 1f;
		} else if (e.getKeyChar() == '/') {
			guiSpeedFactor = 0.03f;
		} else if (integer > 0) {
			guiSpeedFactor = integer;
		} else {
			return;
		}
		frameTimeMS = (long) (1000f * timeStep / guiSpeedFactor);
		System.out.println("Gui speed factor " + Util.StringDoubleTwo(guiSpeedFactor));

	}

	@Override
	public void keyReleased(KeyEvent e) {

	}

	public void takeScreenShot() {
		if(!SimulatorOptions.takeScreenShotsEverySecond)
			return;
		
		Container c = frame.getContentPane();
		BufferedImage im = new BufferedImage(c.getWidth(), c.getHeight(), BufferedImage.TYPE_INT_ARGB);
		c.paint(im.getGraphics());

		String fileName = screenshotFolder + lastImageSecond + ".png";
		try {
			ImageIO.write(im.getSubimage(0, 0, simWidth - 8, frame.getSize().height - 30), "png", new File(fileName));
		} catch (IOException e) {
			e.printStackTrace();
		}
	}
	
	public static void setEnvironmentScreenShotModeOn(){
		SimulatorOptions.onlyShowEnvironment = true;
		SimulatorOptions.scaleGuiToScreen = false;
		SimulatorOptions.drawController = false;
		Gui.scale = 220;
	}
	
	public static void setScreenShotModeOn(){
		Gui.scale = 220;
		SimulatorOptions.scaleGuiToScreen = false;
		SimulatorOptions.drawController = false;
		SimulatorOptions.takeScreenShotsEverySecond = true;
	}
}
