package robot;

import general.Gui;

import java.awt.Color;
import java.awt.Graphics2D;
import java.awt.Polygon;

import joints.ConnectionJoint;






import org.jbox2d.common.Vec2;

import raycasts.GripperRaycastCallback;
import util.Util;

public class Gripper {
	public static final float maxGripperDistance = 0.01f;
	private Robot robot;

	public Gripper(Robot robot) {
		this.robot = robot;
	}

	public boolean tryConnect() {
		for (Robot otherRobot : robot.getWorld().getRobots()) {
			if (!otherRobot.equals(robot) && !robot.getWorld().isJointBetween(robot, otherRobot) && isWithinReach(robot, otherRobot)) {
				
				Vec2[] connectionPoints = getConnectionPoints(otherRobot);
				if (connectionPoints != null) {
					robot.getWorld().addConnectionJoint(new ConnectionJoint(robot, otherRobot, connectionPoints[0], connectionPoints[1]));
					robot.getFrontLight().setColor(Color.YELLOW);
					
					robot.setHasBeenInvolvedInConnection(true);
					otherRobot.setHasBeenInvolvedInConnection(true);
					
					return true;
				}
			}
		}
//		System.out.println("Try connect failed!!!!");
		return false;
	}

	public Vec2[] getConnectionPoints(Robot otherRobot) {
		float angle = robot.getBody().getAngle();

		// From
		Vec2 fromPoint = new Vec2();
		fromPoint.x = (float) (robot.getBody().getPosition().x + Math.cos(angle) * Robot.getSize());
		fromPoint.y = (float) (robot.getBody().getPosition().y + Math.sin(angle) * Robot.getSize());

		// End
		Vec2 endPoint = null;
		try {
			GripperRaycastCallback raycast = new GripperRaycastCallback(otherRobot);
			Vec2 maxEndPoint = new Vec2((float) (maxGripperDistance * Math.cos(angle)), (float) (maxGripperDistance * Math.sin(angle))).add(fromPoint);
			robot.getWorld().raycast(raycast, fromPoint, maxEndPoint);
			endPoint = raycast.getConnectionPoint();
		} catch (Exception e) {
			System.out.println("Exception with gripper raycast");
		}

		if(endPoint == null)
			return null;
		
		Vec2 diff = otherRobot.getBody().getPosition().sub(endPoint);
		float angleOfConnect = Util.getDifferenceBetweenAngles(Util.normalizeAngle(Util.getVec2Angle(diff)) , Util.normalizeAngle(otherRobot.getBody().getAngle()));
		if(angleOfConnect > robot.getLedAngleWidth() / 2){
//			System.out.println("to wide angle of grip, with " + Math.toDegrees(angleOfConnect) + " degrees");
			return null;
		}
		return new Vec2[]{robot.getBody().getLocalPoint(fromPoint), otherRobot.getBody().getLocalPoint(endPoint)};
	}

	public boolean tryDisconnect() {
		if (robot.getConnectionJoint() != null) {
			robot.getWorld().removeConnectionJoint(robot);
			robot.getFrontLight().setColor(Color.BLUE);
			return true;
		}
		return false;
	}

	public boolean isWithinReach(Robot a, Robot b) {
		Vec2 distanceVec = b.getBody().getPosition().sub(a.getBody().getPosition());
		double distance = distanceVec.length() - (2 * Robot.getSize());
	
		return distance < maxGripperDistance;
	}

	public void draw(Graphics2D g) {
		if (robot.isConnected())
			return;

		float angle = robot.getBody().getAngle();
		int x = (int) Math.round(Robot.size * Math.cos(angle) * Gui.scale);
		int y = (int) Math.round(Robot.size * Math.sin(angle) * Gui.scale);

		Polygon polygon = new Polygon();

		float sizeXshort = 0.02f;
		float sizeXLong = 0.04f;

		Vec2[] vertices = new Vec2[4];
		vertices[0] = new Vec2(-sizeXshort / 2, 0f);
		vertices[1] = new Vec2(-sizeXLong / 2, maxGripperDistance);
		vertices[2] = new Vec2(sizeXLong / 2, maxGripperDistance);
		vertices[3] = new Vec2(sizeXshort / 2, 0);

		float rotationAngle = (float) (angle - Math.PI / 2);
		for (Vec2 v : vertices) {
			float nX = (float) (v.x * Math.cos(rotationAngle) - v.y * Math.sin(rotationAngle));
			float nY = (float) (v.x * Math.sin(rotationAngle) + v.y * Math.cos(rotationAngle));
			polygon.addPoint((int) Math.round(nX * Gui.scale + x), (int) Math.round(nY * Gui.scale + y));
		}

		g.setColor(Color.WHITE);
		g.fill(polygon);
		g.setColor(Color.BLACK);
		g.draw(polygon);
	}
}
