package robot;

import environment_objects.Ground;
import environment_objects.World;
import general.Gui;

import java.awt.Color;
import java.awt.Graphics2D;
import java.awt.Polygon;
import java.awt.geom.Ellipse2D;
import java.util.ArrayList;

import objects.BodyObject;
import objects.Light;
import objects.LightSource;
import objects.Wall;
import options.SimulatorOptions;

import org.jbox2d.common.Vec2;

import raycasts.DistanceRaycastCallback;
import raycasts.LightRaycastCallback;

import util.Tuple;
import util.Util;

public class LightSensor extends Sensor {
	// Sensor angle on the robot body
	private float angleOnBody;

	// Sensor direction angle in the world
	private float sensorAngle;

	// Sensor surface in M^2
	private final float sensorSurface = 0.01f * 0.01f;

	// Light intensity in Radiant flux [W]
	private float lightValue;
	private float[] lightColorIntensity = new float[3];

	// Distance in M
	private float distanceValue;

	// Ground value, Sensor distance to ground
	private boolean[] holeDetected;

	// Max angle for detecting objects or light out from sensor
	private float maxDeltaAngle;

	// Max angle for detecting objects or light out from sensor
	private static final float maxMessurableDistance = 0.3f;
	private static final float maxMessurableLightDistance = 0.75f; // Not used // 0.75f;// 2:0.6, 2.5: 0.75, 3: 0.9
	public static final float maxMessurableLightDistanceSquared = maxMessurableLightDistance * maxMessurableLightDistance;// Not used

	public static boolean useRaycast = true;

	private Color closestColorToSensorColor;

	private Robot robot;
	private Vec2 position;
	private World world;

	public static final int HOLE_SENSOR_CLOSE = 0, HOLE_SENSOR_FAR = 1;

	public LightSensor(float angleOnBody, Robot robot, float sensorAngle, World world) {
		this.angleOnBody = angleOnBody;
		this.robot = robot;
		this.maxDeltaAngle = sensorAngle / 2;
		this.world = world;
		position = new Vec2();
	}

	private void resetSensorValues() {
		// Reset value
		lightValue = world.getWorldLightValue();
		distanceValue = Float.MAX_VALUE;
		lightColorIntensity = world.getWorldLightColorIntensity().clone();
		holeDetected = new boolean[] { false, false };

		sensorAngle = Util.normalizeAngle((float) robot.getBody().getAngle() + angleOnBody);
		updatePosition();
	}

	public void readSensor(ArrayList<BodyObject> closeObjects) {
		resetSensorValues();
		for (BodyObject otherObject : closeObjects) {
			// Start right procedure depending on object type
			if (otherObject == robot || otherObject instanceof Ground) {
				continue;
			} else if (otherObject instanceof Robot || otherObject instanceof LightSource) {
				sensingCircleObject(otherObject);
			} else if (otherObject instanceof Wall) {
				sensingWallObject((Wall) otherObject);
			}
		}

		holeDetection();

		if (useRaycast) {
			distanceRaycastOnSensorEdge(maxDeltaAngle);
			distanceRaycastOnSensorEdge(-maxDeltaAngle);
		}

		setClosestColorToSensorColor();
	}

	private void distanceRaycastOnSensorEdge(float relativeAngle) {
		DistanceRaycastCallback raycast = new DistanceRaycastCallback();
		Vec2 point = new Vec2((float) (maxMessurableDistance * Math.cos(relativeAngle + sensorAngle)), (float) (maxMessurableDistance * Math.sin(relativeAngle + sensorAngle))).add(position);
		try {
			world.raycast(raycast, position, point);
			senseDistance(raycast.getDistance());
		} catch (Exception e) {
			System.out.println("Exception with distance raycast " + relativeAngle + " , at position " + position + " pointing at " + point);
		}
	}

	public float getLightValue() {
		return lightValue;
	}

	public float[] getLightColorIntensity() {
		return lightColorIntensity;
	}

	private void setClosestColorToSensorColor() {
		closestColorToSensorColor = Util.getClosestColor(getLightColorIntensity());
	}

	public Color getClosestColorToSensorColor() {
		return closestColorToSensorColor;
	}

	public void senseDistance(float distance) {
		if (distance < distanceValue && distance < maxMessurableDistance)
			distanceValue = Math.max(distance, 0f);
	}

	public float getDistanceValue() {
		return distanceValue;
	}

	private void updatePosition() {
		position.x = (float) (robot.getBody().getPosition().x + Math.cos(sensorAngle) * Robot.getSize());
		position.y = (float) (robot.getBody().getPosition().y + Math.sin(sensorAngle) * Robot.getSize());
	}

	private void senseLight(float valuePerSquareMeter, float[] rgb) {
		// Convert from W/M^2 to W
		float sensesLight = valuePerSquareMeter * sensorSurface;

		lightValue += sensesLight;
		for (int i = 0; i < 3; i++) {
			lightColorIntensity[i] += sensesLight * rgb[i];
		}
	}

	private void sensingCircleObject(BodyObject otherObject) {
		// Distance to other object from sensor
		Vec2 delta = otherObject.getBody().getPosition().sub(position);
		float centerDistance = (float) Math.sqrt(Math.pow(delta.x, 2) + Math.pow(delta.y, 2));

		float edgeDistance = 0;

		if (otherObject instanceof Robot) {
			edgeDistance = centerDistance - Robot.getSize();
		}
		if (otherObject instanceof LightSource) {
			edgeDistance = centerDistance - ((LightSource) otherObject).getSize();
		}

		// Calculate Angles
		float normalAngle = Util.normalizeAngle((float) Math.atan2(delta.y, delta.x));
		float deltaAngle = Util.getDifferenceBetweenAngles(normalAngle, sensorAngle);

		// Update if detected and distance is smaller than previous
		if (deltaAngle < maxDeltaAngle) {
			senseDistance(edgeDistance);
		}
		// Check if object has light
		if (deltaAngle < (float) (Math.PI / 2)) {
			// Find properties of the light of the other object
			Tuple<Light, Float> lightProperties = otherObject.getLightProperties(normalAngle);
			if (lightProperties == null || lightProperties.getF1() == null)
				return;

			Light light = lightProperties.getF1();
			float sourceDirectionFactor = lightProperties.getF2();

			// Check that the source is not behind other objects
			if (useRaycast) {
				try {
					LightRaycastCallback raycast = new LightRaycastCallback(otherObject);
					world.raycast(raycast, position, otherObject.getBody().getPosition());
					if (raycast.isHidden())
						return;
				} catch (Exception e) {
					System.out.println(e.getMessage());
				}
			}

			// Direction factor = 1 if sensor is pointed directly towards light source, decreases with deltaangle
			float lightDirectionFactor = (float) Math.cos(deltaAngle);

			// intensity at distance (per square meters of sensor)
			float intensityAtDistance = (float) (light.getLightPower() / (4 * Math.PI * Math.pow(Math.max(edgeDistance, 0.02), 2)));

			senseLight(lightDirectionFactor * intensityAtDistance * sourceDirectionFactor, light.getRgb());
		}
	}

	private void sensingWallObject(Wall wall) {
		// Sensor
		float edgeDistance;
		float lenghtOfNormalToWall = Util.getNormalBetweenWallAndSensor(wall, position);

		// Check if within
		float leftAngle = Util.normalizeAngle(sensorAngle - maxDeltaAngle);
		float rightAngle = Util.normalizeAngle(sensorAngle + maxDeltaAngle);
		float directionOfWall = wall.getDirectionToWall(position);

		float deltaAngle;

		float angleDeltaWall;

		// special case within
		if (leftAngle > rightAngle) {
			if (directionOfWall == 0.0f) {
				angleDeltaWall = 0.0f;
				deltaAngle = directionOfWall;
			} else if (directionOfWall == (float) (Math.PI / 2.0f)) {
				angleDeltaWall = Util.normalizeAngle(directionOfWall - rightAngle);
				deltaAngle = rightAngle;
			} else if (directionOfWall == (float) (Math.PI)) {
				angleDeltaWall = Util.normalizeAngle(directionOfWall - rightAngle);
				deltaAngle = rightAngle;
			} else {
				angleDeltaWall = Util.normalizeAngle(leftAngle - directionOfWall);
				deltaAngle = leftAngle;
			}
			// System.out.println("Special case");
		}
		// left of sensor
		else if (directionOfWall < leftAngle) {
			float minAngle = Math.min(Util.normalizeAngle(leftAngle - directionOfWall), Util.normalizeAngle(directionOfWall - rightAngle));
			float activeAngle = (Util.normalizeAngle(leftAngle - directionOfWall) < Util.normalizeAngle(directionOfWall - rightAngle)) ? leftAngle : rightAngle;
			angleDeltaWall = minAngle;
			deltaAngle = activeAngle;
		}
		// right of sensor
		else if (directionOfWall > rightAngle) {
			angleDeltaWall = Util.normalizeAngle(directionOfWall - rightAngle);
			deltaAngle = rightAngle;
		} else {
			// within sensor!
			angleDeltaWall = 0.0f;
			deltaAngle = directionOfWall;
		}

		if (angleDeltaWall > (float) (Math.PI / 2)) {
			return;
		}

		edgeDistance = (float) (lenghtOfNormalToWall / Math.cos(angleDeltaWall));
		Vec2 deltaPoint = getDeltaPoint(edgeDistance, deltaAngle, position);

		if (wall.IsPointOnWall(deltaPoint)) {
			senseDistance(edgeDistance);
		}

	}

	private Vec2 getDeltaPoint(float edgeDistance, float deltaAngle, Vec2 sensorPosition) {
		Vec2 deltaPoint = new Vec2();
		deltaPoint.x = (float) (sensorPosition.x + edgeDistance * Math.cos(deltaAngle));
		deltaPoint.y = (float) (sensorPosition.y + edgeDistance * Math.sin(deltaAngle));

		return deltaPoint;
	}

	private void holeDetection() {
		float sensorRangeClose = 0.05f;
		float sensorRangeFar = 0.1f;

		if (world.getGround().getHoleAreas().size() > 0) {
			Vec2 sensorEndPointClose = position.add(Util.createVec2(sensorRangeClose, sensorAngle));
			Vec2 sensorEndPointFar = position.add(Util.createVec2(sensorRangeFar, sensorAngle));

			boolean sensorCloseOverHole = world.getGround().isOverHole(sensorEndPointClose);
			boolean sensorFarOverHole = world.getGround().isOverHole(sensorEndPointFar);

			if (sensorFarOverHole || sensorCloseOverHole) {
				DistanceRaycastCallback raycast = new DistanceRaycastCallback();
				try {
					world.raycast(raycast, position, sensorEndPointFar);
					if (raycast.getDistance() <= sensorRangeFar) {
						holeDetected[HOLE_SENSOR_FAR] = false;
					}
					if (raycast.getDistance() <= sensorRangeClose) {
						holeDetected[HOLE_SENSOR_CLOSE] = false;
					}
				} catch (Exception e) {
					System.out.println("Exception with distance raycast " + 0 + " , at position " + position + " pointing at " + sensorEndPointFar);
				}

				if (sensorCloseOverHole) {
					holeDetected[HOLE_SENSOR_CLOSE] = true;
				}

				if (sensorFarOverHole) {
					holeDetected[HOLE_SENSOR_FAR] = true;
				}
			}
		}
	}

	@Override
	public void draw(Graphics2D g) {
		updatePosition();
		int x = (int) (position.x * Gui.scale);
		int y = (int) (position.y * Gui.scale);

		// Draw sensor point
		float[] normalColor = new float[3];
		float maxColorIntensity = Math.max(Math.max(lightColorIntensity[0], lightColorIntensity[1]), lightColorIntensity[2]);
		// System.out.println(maxColorIntensity);
		maxColorIntensity = Math.max(maxColorIntensity, 0.02f);
		for (int i = 0; i < 3; i++) {
			normalColor[i] = Math.min(Math.max(0f, (float) (lightColorIntensity[i] / maxColorIntensity)), 1f);
		}

		g.setColor(new Color(normalColor[0], normalColor[1], normalColor[2]));
		float size = 3f * Gui.scale / 700;
		Ellipse2D dot = new Ellipse2D.Double(x - size, y - size, size * 2, size * 2);
		g.fill(dot);

		// Draw sensor area
		if (!SimulatorOptions.drawSensors || !robot.getSensorHandler().isDraw())
			return;
		Polygon polygon = new Polygon();
		Vec2[] vertices;

		vertices = new Vec2[20];
		vertices[0] = new Vec2(0f, 0f);

		for (int i = 0; i < 20 - 1; i++) {
			float angle = (float) ((float) i / (20 - 2) * maxDeltaAngle * 2 + robot.getBody().getAngle() + angleOnBody - maxDeltaAngle);
			vertices[i + 1] = new Vec2((float) (maxMessurableDistance * Math.cos(angle)), (float) (maxMessurableDistance * Math.sin(angle)));
		}
		for (Vec2 point : vertices) {
			polygon.addPoint((int) Math.floor(point.x * Gui.scale + x), (int) Math.floor(point.y * Gui.scale + y));
		}

		Color sensorColor = new Color(0, 0, 0, 0.08f);
		g.setColor(sensorColor);
		g.fill(polygon);

		// Draw sensor line
		// g.setStroke(Util.dashedLine);
		// g.draw(polygon);
		// g.setStroke(Util.thinLine);
	}

	public boolean isHoleDetected(int sensorRangeIndex) {
		return holeDetected[sensorRangeIndex];
	}

	public boolean isHoleDetectedInBothRanges() {
		return holeDetected[HOLE_SENSOR_CLOSE] && holeDetected[HOLE_SENSOR_FAR];
	}

	public boolean isHoleDetectedInAnyRange() {
		return holeDetected[HOLE_SENSOR_CLOSE] || holeDetected[HOLE_SENSOR_FAR];
	}

	public String getHoleDetectionOutputString() {
		return (holeDetected[HOLE_SENSOR_CLOSE] ? "Close" : ((holeDetected[HOLE_SENSOR_FAR]) ? "Far" : "False"));
	}
}
