package robot;

import environment_objects.WindArea;
import environment_objects.World;
import general.Gui;
import general.Simulation;

import java.awt.Color;
import java.awt.Graphics2D;
import java.awt.geom.Ellipse2D;

import javax.jws.Oneway;

import joints.AirResistanceJoint;
import joints.ConnectionJoint;
import joints.GroundFrictionJoint;
import objects.BodyObject;
import objects.Light;

import options.SimulatorOptions;

import org.jbox2d.collision.shapes.CircleShape;
import org.jbox2d.common.Vec2;
import org.jbox2d.dynamics.BodyType;
import org.jbox2d.dynamics.FixtureDef;

import util.Tuple;
import util.Util;
import controller.Controller;

public class Robot extends BodyObject {
	/** Constants */

	public static final float size = 0.035f;
	public static final float height = 0.04f;
	public final float maxWheelForce = 0.5f;
	private final float directionFactor = 4f;
	private final float ledAngleWidth = (float) Math.toRadians(70);
	private final float maxVelocity = 0.25f;// Max velocity in M/s

	/** Variables */

	public Ellipse2D ellipse;
	private float forceL, forceR;
	protected boolean motorIsRunning;
	private boolean drawControllerStats = false;
	private Light frontLight, backLight;
	private GroundFrictionJoint frictionJoint;
	private AirResistanceJoint airResistanceJoint;
	private SensorHandler sensorHandler;
	private World world;
	private Controller controller;
	private ConnectionJoint connectionJoint;
	private Gripper gripper;
	private Wheel wheel;
	private float windHitRateFactor = 0;

	private float timeUsed;
	private boolean hasReachedGoalOnce;
	
	private boolean hasBeenInvolvedInConnection = false;

	/** Constructors */

	public Robot(World world, Controller controller) {
		this(new Vec2(), 0f, world, controller);
	}

	public Robot(Vec2 position, float direction, World world, Controller controller) {
		super(position, BodyType.DYNAMIC);
		this.controller = controller;
		this.controller.setup(this);
		this.world = world;
		this.gripper = new Gripper(this);

		bd.angle = (float) Math.toRadians(direction);
		shape = new CircleShape();
		shape.m_radius = size;
		sensorHandler = new SensorHandler(this, world);
		frontLight = new Light(Color.BLUE);
		backLight = new Light(Color.GREEN);
		motorIsRunning = true;
		wheel = new Wheel(this);

		timeUsed = 0f;
		hasReachedGoalOnce = false;
	}

	/** Private Methods */

	private void applyMotorForce() {
		// System.out.println(body.m_linearVelocity.length() / maxVelocity);

		if (forceR == 0 && forceL == 0)
			return;

		// Calculate angular force
		float angularForce = (forceL - forceR) * directionFactor;

		// Apply forces
		wheel.applyWheelForce();
		body.setAngularVelocity(angularForce);
	}

	private void applyWind() {
		WindArea windArea = world.getGround().getWindArea(getBody().getPosition());
		if (windArea == null) {
			windHitRateFactor = 0;
			return;
		}
		windHitRateFactor = windArea.getWindHitRateFactor(getBody().getPosition(), size);

		Vec2 windOnRobot = windArea.getWindForce().mul(windHitRateFactor);
		body.applyForceToCenter(windOnRobot);
	}

	private void handleHoleFalling() {
		Vec2 hole = world.getGround().getCloseHole(getBody().getPosition());
		if (hole != null) {
			motorIsRunning = false;
			drawControllerStats = false;
			Vec2 diff = hole.sub(getBody().getPosition());
			float dir = Util.getVec2Angle(diff);
			Vec2 force = Util.createVec2(15, dir);
			body.applyForceToCenter(force);
			world.removeObject(this);
		}
	}

	/** Public Methods */

	public void run(float elapsedTime) {
		sensorHandler.readSensors();
		controller.run(elapsedTime);
	}

	@Override
	public Tuple<Light, Float> getLightProperties(float normalAngle) {
		Light light;
		float sourceDirectionFactor = 1f;

		float otherAngleFront = Util.normalizeAngle(getBody().getAngle());
		float deltaSensePointFront = Util.getDifferenceBetweenAngles(otherAngleFront, normalAngle);

		float otherAngleBack = Util.normalizeAngle((float) (otherAngleFront + Math.PI));
		float deltaSensePointBack = Util.getDifferenceBetweenAngles(otherAngleBack, normalAngle);

		float delteSensePoint;
		if (deltaSensePointFront < deltaSensePointBack) {
			light = getBackLight();
			delteSensePoint = deltaSensePointFront;
		} else {
			light = getFrontLight();
			delteSensePoint = deltaSensePointBack;
		}

		if (delteSensePoint > getLedAngleWidth() / 2) {
			float diff = delteSensePoint - getLedAngleWidth() / 2;
			float maxSensAngle = (float) (Math.PI / 15);
			// System.out.println(Math.toDegrees(maxSensAngle));
			sourceDirectionFactor = Math.max((maxSensAngle - diff) / maxSensAngle, 0);
		}
		return new Tuple<Light, Float>(light, sourceDirectionFactor);
	}

	public boolean isConnected() {
		return connectionJoint != null;
	}

	@Override
	public void update() {
		// Update joints
		frictionJoint.update(getBody().getPosition(), body.m_linearVelocity.length() >= 0.01);
		airResistanceJoint.update(getBody().getLinearVelocity());

		// Handle hole falling
		handleHoleFalling();

		// Apply motor force to the wheels
		if (!motorIsRunning) {
			forceL = 0;
			forceR = 0;
		}

		wheel.updateWheelForce(forceL, forceR);
		applyMotorForce();

		// Apply windforce
		applyWind();
	}

	public boolean hasReachedGoalOnce() {
		return hasReachedGoalOnce;
	}

	@Override
	public void draw(Graphics2D g) {
		if(SimulatorOptions.onlyShowEnvironment)
			return;
		double scaledRadius = shape.m_radius * Gui.scale;
		ellipse = new Ellipse2D.Double(-scaledRadius, -scaledRadius, 2 * scaledRadius, 2 * scaledRadius);
		float x = body.getPosition().x * Gui.scale;
		float y = body.getPosition().y * Gui.scale;
		int dirY = (int) Math.round(scaledRadius * Math.sin(body.getAngle()));
		int dirX = (int) Math.round(scaledRadius * Math.cos(body.getAngle()));
		g.translate(x, y);

		// Body
		if (drawControllerStats) {
			g.setColor(Color.YELLOW);
		} else {
			g.setColor(Color.WHITE);
		}
		g.fill(ellipse);

		// Rest of body outline
		g.setStroke(Util.veryThinLine);
		g.setColor(Color.BLACK);
		g.draw(ellipse);

		// Direction
		g.setStroke(Util.thinLine);
		g.setColor(Color.BLACK);
		g.drawLine(0, 0, dirX, dirY);
		gripper.draw(g);

		// Back leds
		g.setStroke(Util.mediumLine);
		g.setColor(backLight.getColor());
		g.drawArc(-(int) (size * Gui.scale), -(int) (size * Gui.scale), (int) (2 * size * Gui.scale), (int) (2 * size * Gui.scale), (int) Math.round(Math.toDegrees(-getBody().getAngle() + Math.PI - ledAngleWidth / 2)),
				(int) Math.toDegrees(ledAngleWidth));

		// Front leds
		g.setColor(frontLight.getColor());
		g.drawArc(-(int) (size * Gui.scale), -(int) (size * Gui.scale), (int) (2 * size * Gui.scale), (int) (2 * size * Gui.scale), (int) Math.round(Math.toDegrees(-getBody().getAngle() - ledAngleWidth / 2)), (int) Math.toDegrees(ledAngleWidth));

		// Draw wind percent
		g.setColor(Color.BLUE);
		if (SimulatorOptions.drawWindPercent) {
			String wind = Math.round(windHitRateFactor * 100) + "%";
			g.drawString(wind, -Util.getStringWithOnScreen(g, wind) / 2, size * Gui.scale * 0.5f);
		}

		g.translate(-x, -y);
		// Sensors
		sensorHandler.draw(g);
	}

	/**
	 * Getters and Setters
	 */

	public Light getFrontLight() {
		return frontLight;
	}

	public Light getBackLight() {
		return backLight;
	}

	public Controller getController() {
		return controller;
	}

	public void setDrawController(boolean drawController) {
		this.drawControllerStats = drawController;
	}

	public static final float getSize() {
		return size;
	}

	public GroundFrictionJoint getFrictionJoint() {
		return frictionJoint;
	}

	public void setFrictionJoint(GroundFrictionJoint frictionJoint) {
		this.frictionJoint = frictionJoint;
	}

	public ConnectionJoint getConnectionJoint() {
		return connectionJoint;
	}

	public void setConnectionJoint(ConnectionJoint connectionJoint) {
		this.connectionJoint = connectionJoint;
	}

	public void setWheelForces(float forceLeft, float forceRight) {
		forceL = Math.max(-1, Math.min(1, forceLeft));
		forceR = Math.max(-1, Math.min(1, forceRight));
	}

	public Gripper getGripper() {
		return gripper;
	}

	public World getWorld() {
		return world;
	}

	public boolean getMotorIsRunning() {
		return motorIsRunning;
	}

	public void setMotorIsRunning(boolean motorIsRunning) {
		this.motorIsRunning = motorIsRunning;
	}

	public float getLedAngleWidth() {
		return ledAngleWidth;
	}

	public FixtureDef setFixtureDef() {
		FixtureDef fd = new FixtureDef();
		fd.shape = shape;
		fd.density = 52.0f;
		fd.friction = 0.1f;
		fd.restitution = 0.1f;
		return fd;
	}

	public SensorHandler getSensorHandler() {
		return sensorHandler;
	}

	public void setAirResistanceJoint(AirResistanceJoint airResistanceJoint) {
		this.airResistanceJoint = airResistanceJoint;
	}

	public float getTimeUsed() {
		return timeUsed;
	}

	public void setTimeUsed(float timeUsed) {
		this.timeUsed = timeUsed;
	}

	public void setHasReachedGoalOnce(boolean hasReachedGoalOnce) {
		this.hasReachedGoalOnce = hasReachedGoalOnce;
	}

	@Override
	public boolean isOccupied(Vec2 position) {
		float dX = getBody().getPosition().x - position.x;
		float dY = getBody().getPosition().y - position.y;
		float l = (float) (Math.sqrt(Math.pow(dX, 2) + Math.pow(dY, 2)));
		// System.out.println(radius > l);
		return size > l;
	}

	public float getForceL() {
		return forceL;
	}

	public float getForceR() {
		return forceR;
	}

	public boolean hasBeenInvolvedInConnection() {
		return hasBeenInvolvedInConnection;
	}

	public void setHasBeenInvolvedInConnection(boolean hasBeenInvolvedInConnection) {
		this.hasBeenInvolvedInConnection = hasBeenInvolvedInConnection;
	}
}
