package robot;

import environment_objects.World;

import java.awt.Graphics2D;
import java.util.ArrayList;

import options.SimulatorOptions;

import org.jbox2d.common.Vec2;

import objects.BodyObject;
import objects.Wall;

import robot.LightSensor;
import robot.Robot;
import robot.Sensor;

public class SensorHandler {
	private ArrayList<Sensor> sensors;
	private ArrayList<LightSensor> lightSensors;
	private World world;
	private Robot robot;
	private Accelerometer accelerometer;
	private boolean draw = true;

	public static final int FRONT = 0, FRONT_RIGHT = 1, RIGHT = 2, BACK = 3, LEFT = 4, FRONT_LEFT = 5;

	public SensorHandler(Robot robot, World world) {
		float sensorAngleWidth = (float) (Math.toRadians(45));
		sensors = new ArrayList<Sensor>();
		lightSensors = new ArrayList<LightSensor>();
		this.world = world;
		this.robot = robot;
		addSensor(new LightSensor(0f, robot, sensorAngleWidth, world));
		addSensor(new LightSensor((float) Math.toRadians(45), robot, sensorAngleWidth, world));
		addSensor(new LightSensor((float) Math.toRadians(90), robot, sensorAngleWidth, world));
		addSensor(new LightSensor((float) Math.toRadians(180), robot, sensorAngleWidth, world));
		addSensor(new LightSensor((float) Math.toRadians(270), robot, sensorAngleWidth, world));
		addSensor(new LightSensor((float) Math.toRadians(315), robot, sensorAngleWidth, world));
		addSensor(new Accelerometer(robot));
	}

	private void addSensor(Sensor sensor) {
		sensors.add(sensor);

		if (sensor instanceof LightSensor) {
			lightSensors.add((LightSensor) sensor);
		} else if (sensor instanceof Accelerometer) {
			accelerometer = (Accelerometer) sensor;
		}
	}

	public void readSensors() {
		// find close objects, returns all if useMaxLightSensingDistance == true
		ArrayList<BodyObject> closeObjects = findCloseObjects();

		for (LightSensor s : lightSensors) {
			s.readSensor(closeObjects);
		}

		accelerometer.readSensor();
	}

	public ArrayList<BodyObject> findCloseObjects() {
		if (!SimulatorOptions.useMaxLightSensingDistance)
			return world.getWorldObjects();

		ArrayList<BodyObject> closeObjects = new ArrayList<BodyObject>();

		for (BodyObject object : world.getWorldObjects()) {
			if (object instanceof Wall) {
				closeObjects.add(object);
				continue;
			}
			Vec2 delta = object.getBody().getPosition().sub(robot.getBody().getPosition());
			float squareDistance = (float) (Math.pow(delta.x, 2) + Math.pow(delta.y, 2));
			if (squareDistance < LightSensor.maxMessurableLightDistanceSquared) {
				closeObjects.add(object);
			}
		}

		return closeObjects;
	}

	public void printDistances() {
		System.out.println();
		for (Sensor s : sensors) {
			if (s instanceof LightSensor) {
				System.out.print(((LightSensor) s).getDistanceValue() + "\t\t");
			}
		}
	}

	public void printLightSensorValues() {
		System.out.println();
		for (Sensor s : sensors) {
			if (s instanceof LightSensor) {
				System.out.print(((LightSensor) s).getLightValue() + "\t\t");
			}
		}
	}

	public void draw(Graphics2D g) {
		for (Sensor s : sensors) {
			s.draw(g);
		}
	}

	public ArrayList<Sensor> getSensors() {
		return sensors;
	}

	public ArrayList<LightSensor> getLightSensors() {
		return lightSensors;
	}

	public Accelerometer getAccelerometer() {
		return accelerometer;
	}

	public void setDraw(boolean draw) {
		this.draw = draw;
	}

	public boolean isDraw() {
		return draw;
	}
}
