package robot;

import general.Simulation;

import org.jbox2d.common.Vec2;

import util.Util;

public class Wheel {
	private float wheelForce;
	private Robot robot;
	private Vec2 oldGlobalPosition;
	private final float wheelRadius = 0.025f;
	private final float gearRatio = 1 / 50f; // number from epuck spec: 50:1
	private final float efficiency = 1f; // 100%
	private final float rollingFriction = 0.001f; // a typical rolling friction from wikipedia
	
	public Wheel(Robot robot) {
		this.robot = robot;
		oldGlobalPosition = null;
	}

	public void updateWheelForce(float forcePercentL, float forcePercentR) {
		float forcePercent = (forcePercentL+forcePercentR)/2;
		
		if (oldGlobalPosition == null) {
			oldGlobalPosition = robot.getBody().getPosition().clone();
		}
		
		float maxTorque = Math.min(Math.abs(forcePercent), 1f) * robot.getBody().getMass() * 9.81f * robot.getFrictionJoint().getFriction();
		float freeSpeedMotor = 4829 * Math.min(Math.abs(forcePercent), 1f); // to obtain a maxSpeed of 0.25 m/s, we need a freeMotorSpeed of 4829 rpm. (epuck: 20 steps/revolution, 18 degrees resolution = 50 rps = 3000 rpm)

		Vec2 newPosition = robot.getBody().getPosition().clone();

		float speedRobot = newPosition.sub(oldGlobalPosition).length() / Simulation.timeStep;
		
//		System.out.println("speeds: new: " + speedRobot + ", old: " + robot.getBody().getLinearVelocity().length() + ", force%: " + forcePercent);
		speedRobot = Math.min(speedRobot, robot.getBody().getLinearVelocity().length());

		float outputRPM = speedRobot * 60 / (float) (wheelRadius * 2 * Math.PI);
		float motorRPM = Math.min(freeSpeedMotor, outputRPM / gearRatio);

		float motorTorque = -(maxTorque / freeSpeedMotor) * motorRPM + maxTorque;
		float gearboxTorque = motorTorque * gearRatio * efficiency;

		float accelerationForce = gearboxTorque / wheelRadius;

		// rolling friction
		float frictionResistance = robot.getBody().getMass() * 9.81f * rollingFriction;

		if (accelerationForce > 0) {
			wheelForce = accelerationForce - frictionResistance;
			if (wheelForce < 0) {
				wheelForce = 0;
			}
		} else {
			wheelForce = 0;
		}

		float a = wheelForce / robot.getBody().getMass();

		if (forcePercent < 0) {
			a = -a;
		}

//		System.out.println("a: " + a);
//		System.out.println("WheelF: " + wheelForce + ", v body: " + robot.getBody().getLinearVelocity().length() + ", speedRobot: " + speedRobot + ", motorRPM: " + motorRPM + ", maxTorque: " + maxTorque);

		oldGlobalPosition = robot.getBody().getPosition().clone();

		// because of rolling, we need to counteract sliding friction
		float slidingFriction = robot.getFrictionJoint().getJoint().getMaxForce();
		wheelForce += slidingFriction;

		if (forcePercent < 0) {
			wheelForce = -wheelForce;
		}
		
//		System.out.println("WheelForce final: " + wheelForce);
	}

	public void applyWheelForce() {
		Vec2 motorForceVector = Util.createVec2(wheelForce, robot.getBody().getAngle());
		robot.getBody().applyForceToCenter(motorForceVector);
//		System.out.println("MotorForceVector: " + motorForceVector);
	}

	public float getWheelForce() {
		return wheelForce;
	}

}
