package setups;

import environment_objects.Hole;
import environment_objects.IceCircle;
import environment_objects.IceRectangle;
import environment_objects.WindArea;
import environment_objects.World;
import general.DistributeRandomly;
import general.Phenotype;

import java.awt.Color;
import java.io.File;
import java.io.IOException;

import objects.LightSource;
import options.SimulatorOptions;

import org.ini4j.Ini;
import org.ini4j.InvalidFileFormatException;
import org.jbox2d.common.Vec2;

import robot.Robot;
import controller.ArbitrationController;

public class SetupFromWorldFile extends Setup {

	private Ini ini;

	private String filename;

	// World
	private float worldWidth, worldHeight;

	// StartingArea
	private float startingAreaX, startingAreaY, startingAreaDx, startingAreaDy;

	// Robot
	private int numberOfRobots;

	// LightSource
	private float[] lightSourceX, lightSourceY, lightSourceGoal;
	private Color[] colors;

	// IceAreaRectangle
	private float[] iceAreaRectangleX, iceAreaRectangleY, iceAreaRectangleSizeX, iceAreaRectangleSizeY;

	// IceAreaCircle
	private float[] iceAreaCircleX, iceAreaCircleY, iceAreaCircleRadius;

	// WindArea
	private float[] windAreaX, windAreaY, windAreaSizeX, windAreaSizeY, windAreaVelocity;
	private int[] windAreaDirection;

	// HoleArea
	private float[] holeAreaX, holeAreaY, holeAreaRadius;

	public SetupFromWorldFile(String filename) {
		this.filename = filename;
		ini = readFromFile();
		saveAllData();
	}

	@SuppressWarnings("unchecked")
	@Override
	public World generateWorld(Phenotype<?> phenotype) {
		if(phenotype != null){
			phenotype.setNumberOfRobotsPerSimulation(numberOfRobots);
		}
		
		if (ini == null) {
			System.out.println("Null i setupFromFile");
			return null;
		}

		World world = new World(worldWidth, worldHeight);

		world.getGround().setStartingArea(startingAreaX, startingAreaY, startingAreaDx, startingAreaDy);

		for (int i = 0; i < numberOfRobots; i++) {
			world.addObject(new Robot(world, new ArbitrationController((Phenotype<Float>) phenotype)));
		}

		DistributeRandomly.distributeRobotsRandomly(world);

		if (lightSourceX != null && lightSourceY != null && lightSourceGoal != null && colors != null) {
			for (int i = 0; i < lightSourceX.length; i++) {
				LightSource lightSource = new LightSource(new Vec2(lightSourceX[i], lightSourceY[i]), colors[i]);
				if (lightSourceGoal[i] == 1) {
					world.setGoalObject(lightSource);
				}
				world.addObject(lightSource);
			}
		}

		if (iceAreaRectangleX != null && iceAreaRectangleY != null && iceAreaRectangleSizeX != null && iceAreaRectangleSizeY != null) {
			for (int i = 0; i < iceAreaRectangleX.length; i++) {
				world.getGround().addIceArea(new IceRectangle(new Vec2(iceAreaRectangleX[i], iceAreaRectangleY[i]), new Vec2(iceAreaRectangleSizeX[i], iceAreaRectangleSizeY[i])));
			}
		}

		if (iceAreaCircleX != null && iceAreaCircleY != null && iceAreaCircleRadius != null) {
			for (int i = 0; i < iceAreaCircleX.length; i++) {
				world.getGround().addIceArea(new IceCircle(new Vec2(iceAreaCircleX[i], iceAreaCircleY[i]), iceAreaCircleRadius[i]));
			}
		}

		if (windAreaX != null && windAreaY != null && windAreaSizeX != null && windAreaSizeY != null && windAreaVelocity != null && windAreaDirection != null) {
			for (int i = 0; i < windAreaX.length; i++) {
				world.getGround().addWindArea(new WindArea(new Vec2(windAreaX[i], windAreaY[i]), new Vec2(windAreaSizeX[i], windAreaSizeY[i]), windAreaVelocity[i], windAreaDirection[i], world));
			}
		}

		if (holeAreaX != null && holeAreaY != null && holeAreaRadius != null) {
			for (int i = 0; i < holeAreaX.length; i++) {
				world.getGround().addHole(new Hole(new Vec2(holeAreaX[i], holeAreaY[i]), holeAreaRadius[i]));
			}
		}

		SimulatorOptions.drawController = true;

		if (ini == null) {
			return null;
		}

		return world;

	}

	public Ini readFromFile() {
		try {
			Ini ini = new Ini(new File(SimulatorOptions.environmentsFolder + filename));
			return ini;

		} catch (InvalidFileFormatException e) {
			e.printStackTrace();
		} catch (IOException e) {
			e.printStackTrace();
		}
		return null;

	}

	public void saveAllData() {
		if (ini == null) {
			return;
		}

		Ini.Section worldSection = ini.get("World");
		if (worldSection != null) {
			worldWidth = worldSection.get("width", float.class);
			worldHeight = worldSection.get("height", float.class);
		}

		Ini.Section startingareaSection = ini.get("StartingArea");
		if (startingareaSection != null) {
			startingAreaX = startingareaSection.get("x", float.class);
			startingAreaY = startingareaSection.get("y", float.class);
			startingAreaDx = startingareaSection.get("dx", float.class);
			startingAreaDy = startingareaSection.get("dy", float.class);
		}

		Ini.Section robotSection = ini.get("Robot");
		if (robotSection != null) {
			numberOfRobots = robotSection.get("nmbOfRobots", int.class);
		}

		Ini.Section lightSourceSection = ini.get("LightSource");
		if (lightSourceSection != null) {
			lightSourceX = lightSourceSection.getAll("x", float[].class);
			lightSourceY = lightSourceSection.getAll("y", float[].class);
			lightSourceGoal = lightSourceSection.getAll("goal", float[].class);
			float[] redColors = lightSourceSection.getAll("r", float[].class);
			float[] greenColors = lightSourceSection.getAll("g", float[].class);
			float[] blueColors = lightSourceSection.getAll("b", float[].class);

			colors = new Color[redColors.length];

			for (int i = 0; i < colors.length; i++) {
				colors[i] = new Color(redColors[i], greenColors[i], blueColors[i]);
			}
		}

		Ini.Section iceAreaRectangleSection = ini.get("IceAreaRectangle");
		if (iceAreaRectangleSection != null) {
			iceAreaRectangleX = iceAreaRectangleSection.getAll("x", float[].class);
			iceAreaRectangleY = iceAreaRectangleSection.getAll("y", float[].class);
			iceAreaRectangleSizeX = iceAreaRectangleSection.getAll("sizeX", float[].class);
			iceAreaRectangleSizeY = iceAreaRectangleSection.getAll("sizeY", float[].class);
		}

		Ini.Section iceAreaCircleSection = ini.get("IceAreaCircle");
		if (iceAreaCircleSection != null) {
			iceAreaCircleX = iceAreaCircleSection.getAll("x", float[].class);
			iceAreaCircleY = iceAreaCircleSection.getAll("y", float[].class);
			iceAreaCircleRadius = iceAreaCircleSection.getAll("radius", float[].class);
		}

		Ini.Section windAreaSection = ini.get("WindArea");
		if (windAreaSection != null) {
			windAreaX = windAreaSection.getAll("x", float[].class);
			windAreaY = windAreaSection.getAll("y", float[].class);
			windAreaSizeX = windAreaSection.getAll("sizeX", float[].class);
			windAreaSizeY = windAreaSection.getAll("sizeY", float[].class);
			windAreaVelocity = windAreaSection.getAll("velocity", float[].class);
			windAreaDirection = windAreaSection.getAll("direction", int[].class);
		}

		Ini.Section holeAreaSection = ini.get("HoleArea");
		if (holeAreaSection != null) {
			holeAreaX = holeAreaSection.getAll("x", float[].class);
			holeAreaY = holeAreaSection.getAll("y", float[].class);
			holeAreaRadius = holeAreaSection.getAll("radius", float[].class);
		}

	}
}
